"""The Rule class defines a rule for determining the next question presented to an agent.

The key component is an expression specifiying the logic of the rule, which can include any combination of logical operators ('and', 'or', 'not'), e.g.:

.. code-block:: python

    "q1 == 'yes' or q2 == 'no'"

The expression must be about questions "before" the current question.

Only one rule should apply at each priority level.
If multiple rules apply, the one with the highest priority is used. 
If there are conflicting rules, an exception is raised.

If no rule is specified, the next question is given as the default.
When a question is added with index, it is always given a rule the next question is index + 1, but
with a low (-1) priority.
"""
import ast
from collections import namedtuple
from rich import print
from simpleeval import EvalWithCompoundTypes
from typing import Any, Union, List
from edsl.exceptions import (
    SurveyRuleCannotEvaluateError,
    SurveyRuleCollectionHasNoRulesAtNodeError,
    SurveyRuleRefersToFutureStateError,
    SurveyRuleReferenceInRuleToUnknownQuestionError,
    SurveyRuleSendsYouBackwardsError,
    SurveyRuleSkipLogicSyntaxError,
)
from edsl.surveys.base import EndOfSurvey
from edsl.utilities.ast_utilities import extract_variable_names
from edsl.utilities.interface import print_table_with_rich


class Rule:
    """The Rule class defines a "rule" for determining the next question presented to an agent."""

    # Not implemented but nice to have:
    # We could potentially use the question pydantic models to check for rule conflicts, as
    # they define the potential trees through a survey.

    # We could also use the AST to check for conflicts by inspecting the types of a rule.
    # For example, if we know the answer to a question is a string, we could check that
    # the expression only contains string comparisons.

    def __init__(
        self,
        current_q: int,
        expression: str,
        next_q: Union[int, EndOfSurvey.__class__],
        question_name_to_index: dict[str, int],
        priority: int,
    ):
        """Represent a rule for determining the next question presented to an agent.

        Questions are represented by int indices.

        :param current_q: The question at which the rule is potentially applied.
        :param expression: A string that evaluates to true or false. If true, then next_q is next.
        :param next_q: The next question if the expression is true.
        :param question_name_to_index: A dictionary mapping question names to indices.
        :param priority: An integer that determines which rule is applied, if multiple rules apply.
        """
        self.current_q = current_q
        self.expression = expression
        self.next_q = next_q
        self.question_name_to_index = question_name_to_index
        self.priority = priority

        if not next_q == EndOfSurvey and current_q > next_q:
            raise SurveyRuleSendsYouBackwardsError

        # get the AST for the expression - used to extract
        # the variables referenced in the expression
        try:
            self.ast_tree = ast.parse(self.expression)
        except SyntaxError:
            raise SurveyRuleSkipLogicSyntaxError

        # get the names of the variables in the expression
        # e.g., q1 == 'yes' -> ['q1']
        extracted_question_names = extract_variable_names(self.ast_tree)

        # make sure all the variables in the expression are known questions
        try:
            assert all([q in question_name_to_index for q in extracted_question_names])
        except AssertionError:
            print(f"Question name to index: {question_name_to_index}")
            print(f"Extracted question names: {extracted_question_names}")
            raise SurveyRuleReferenceInRuleToUnknownQuestionError

        # get the indices of the questions mentioned in the expression
        self.named_questions_by_index = [
            question_name_to_index[q] for q in extracted_question_names
        ]

        # A rule should only refer to questions that have already been asked.
        # so the named questions in the expression should not be higher than the current question
        if self.named_questions_by_index:
            if max(self.named_questions_by_index) > self.current_q:
                raise SurveyRuleRefersToFutureStateError

    def to_dict(self):
        """Convert the rule to a dictionary for serialization."""
        return {
            "current_q": self.current_q,
            "expression": self.expression,
            "next_q": "EndOfSurvey" if self.next_q == EndOfSurvey else self.next_q,
            "priority": self.priority,
            "question_name_to_index": self.question_name_to_index,
        }

    @classmethod
    def from_dict(self, rule_dict):
        """Create a rule from a dictionary."""
        if rule_dict["next_q"] == "EndOfSurvey":
            rule_dict["next_q"] = EndOfSurvey

        return Rule(
            current_q=rule_dict["current_q"],
            expression=rule_dict["expression"],
            next_q=rule_dict["next_q"],
            priority=rule_dict["priority"],
            question_name_to_index=rule_dict["question_name_to_index"],
        )

    def __repr__(self):
        """Pretty-print the rule."""
        return f'Rule(current_q={self.current_q}, expression="{self.expression}", next_q={self.next_q}, priority={self.priority}, question_name_to_index={self.question_name_to_index})'

    def __str__(self):
        """Return a string representation of the rule."""
        return self.__repr__()

    @property
    def question_index_to_name(self):
        """Reverse the dictionary do we can look up questions by name."""
        return {v: k for k, v in self.question_name_to_index.items()}

    def show_ast_tree(self):
        """Pretty-print the AST tree to the terminal."""
        print(
            ast.dump(
                self.ast_tree, annotate_fields=True, indent=4, include_attributes=True
            )
        )

    def evaluate(self, answers: dict[int, Any]):
        """Compute the value of the expression, given a dictionary of known questions answers.

        If the expression cannot be evaluated, it raises a CannotEvaluate exception.
        """

        def substitute_in_answers(expression, answers):
            """Take the dictionary of answers and substitute them into the expression."""
            for var, value in answers.items():
                # If it is a string, add quotes; otherwise, just convert to string.
                if isinstance(value, str):
                    replacement = f"'{value}'"
                else:
                    replacement = str(value)

                expression = expression.replace(var, replacement)
            return expression

        try:
            to_evaluate = substitute_in_answers(self.expression, answers)
            return EvalWithCompoundTypes().eval(to_evaluate)
        except Exception as e:
            print(f"Exception in evaluation: {e}")
            raise SurveyRuleCannotEvaluateError


if __name__ == "__main__":
    r = Rule(
        current_q=1,
        expression="q1 == 'yes'",
        next_q=2,
        question_name_to_index={"q1": 1},
        priority=0,
    )



import asyncio
from enum import Enum
from typing import Literal, List, Type, DefaultDict
from collections import UserDict, defaultdict

from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.enums import pricing, TokenPricing
from edsl.jobs.tasks.task_status_enum import TaskStatus

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]

from edsl.jobs.interviews.InterviewStatistic import InterviewStatistic
from edsl.jobs.interviews.InterviewStatisticsCollection import (
    InterviewStatisticsCollection,
)


class JobsRunnerStatusData:
    pricing = pricing

    def status_dict(self, interviews):
        status = []
        for interview in interviews:
            # model = interview.model
            status.append(interview.interview_status)

        return status
        # return model_to_status

    def status_counts(self, interviews):
        model_to_status = defaultdict(InterviewStatusDictionary)

        for interview in interviews:
            model = interview.model
            model_to_status[model] += interview.interview_status

        # breakpoint()
        return model_to_status.values()

    def generate_status_summary(
        self,
        completed_tasks: List[Type[asyncio.Task]],
        elapsed_time: float,
        interviews: List[Type["Interview"]],
    ) -> InterviewStatisticsCollection:
        """Generate a summary of the status of the job runner.

        :param completed_tasks: list of completed tasks
        :param elapsed_time: time elapsed since the start of the job
        :param interviews: list of interviews to be conducted
        """

        models_to_tokens = defaultdict(InterviewTokenUsage)
        model_to_status = defaultdict(InterviewStatusDictionary)

        waiting_dict = defaultdict(int)

        interview_statistics = InterviewStatisticsCollection()

        for interview in interviews:
            model = interview.model
            models_to_tokens[model] += interview.token_usage
            model_to_status[model] += interview.interview_status
            waiting_dict[model] += interview.interview_status.waiting

        interview_statistics.add_stat(
            InterviewStatistic(
                "elapsed_time", value=elapsed_time, digits=1, units="sec."
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "total_interviews_requested", value=len(interviews), units=""
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "completed_interviews", value=len(completed_tasks), units=""
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "percent_complete",
                value=len(completed_tasks) / len(interviews) * 100
                if len(interviews) > 0
                else "NA",
                digits=0,
                units="%",
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "average_time_per_interview",
                value=elapsed_time / len(completed_tasks) if completed_tasks else "NA",
                digits=1,
                units="sec.",
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "task_remaining", value=len(interviews) - len(completed_tasks), units=""
            )
        )
        number_remaining = len(interviews) - len(completed_tasks)
        time_per_task = (
            elapsed_time / len(completed_tasks) if len(completed_tasks) > 0 else "NA"
        )
        estimated_time_remaining = (
            number_remaining * time_per_task if time_per_task != "NA" else "NA"
        )

        interview_statistics.add_stat(
            InterviewStatistic(
                "estimated_time_remaining",
                value=estimated_time_remaining,
                digits=1,
                units="sec.",
            )
        )
        model_queues_info = []
        for model, num_waiting in waiting_dict.items():
            model_info = self._get_model_info(model, num_waiting, models_to_tokens)
            model_queues_info.append(model_info)

        interview_statistics["model_queues"] = model_queues_info

        return interview_statistics

    def _get_model_info(
        self,
        model: str,
        num_waiting: int,
        models_to_tokens: InterviewTokenUsageMapping,
    ):
        """Get the status of a model."""
        if model.model not in self.pricing:
            raise ValueError(f"Model {model.model} not found in pricing")

        prices = self.pricing[model.model]

        model_info = {
            "model_name": model.model,
            "TPM_limit_k": model.TPM / 1000,
            "RPM_limit_k": model.RPM / 1000,
            "num_tasks_waiting": num_waiting,
            "token_usage_info": [],
        }

        token_usage_types = ["new_token_usage", "cached_token_usage"]
        for token_usage_type in token_usage_types:
            cache_info = self._get_token_usage_info(
                token_usage_type, models_to_tokens, model, prices
            )
            model_info["token_usage_info"].append(cache_info)

        return model_info

    def _get_token_usage_info(
        self,
        cache_status: Literal["new_token_usage", "cached_token_usage"],
        models_to_tokens: InterviewTokenUsageMapping,
        model: str,
        prices: TokenPricing,
    ):
        cache_info = {"cache_status": cache_status, "details": []}
        token_usage = getattr(models_to_tokens[model], cache_status)
        for token_type in ["prompt_tokens", "completion_tokens"]:
            tokens = getattr(token_usage, token_type)
            cache_info["details"].append(
                {
                    "type": token_type,
                    "tokens": tokens,
                }
            )
        cache_info["cost"] = f"${token_usage.cost(prices):.5f}"
        return cache_info



from edsl.language_models.DeepInfra import create_deep_infra_model
from edsl.enums import LanguageModelType

model_name = LanguageModelType.LLAMA_2_13B_CHAT_HF.value
url = "https://api.deepinfra.com/v1/inference/meta-llama/Llama-2-13b-chat-hf"
#url = "https://api.deepinfra.com/v1/inference/meta-llama/Llama-2-70b-chat-hf"
LlamaTwo13B = create_deep_infra_model(
    model_name=model_name, url=url, model_class_name="LlamaTwo13B"
)

if __name__ == "__main__":
    model = LlamaTwo13B(use_cache=False)
    from edsl.questions import QuestionMultipleChoice

    q = QuestionMultipleChoice(
        question_text="Are pickled pigs feet a popular breakfast food in the US?",
        question_options=["Yes", "No", "Unsure"],
        question_name="bkfast_question",
    )
    results = q.by(model).run()
    results.select("answer.*", "model.model").print()



"""Agent persona for a human agent."""
import textwrap

from edsl.prompts.Prompt import PromptBase
from edsl.prompts.prompt_config import ComponentTypes

from edsl.enums import LanguageModelType


class AgentPersona(PromptBase):
    """Agent persona for a human agent."""

    model = "gpt-4-1106-preview"
    component_type = ComponentTypes.AGENT_PERSONA
    default_instructions = textwrap.dedent(
        """\
            You are an agent with the following persona:
            {{ traits }}
        """
    )



"""This module contains the LanguageModel class, which is an abstract base class for all language models."""
from __future__ import annotations
import warnings
from functools import wraps
import asyncio
import json
import time
import inspect
import os

from typing import Coroutine, Any, Callable, Type, List, get_type_hints

from abc import ABC, abstractmethod, ABCMeta

from rich.table import Table

from edsl.language_models.schemas import model_prices
from edsl.utilities.decorators import sync_wrapper, jupyter_nb_handler
from edsl.language_models.repair import repair
from edsl.exceptions.language_models import LanguageModelAttributeTypeError
from edsl.enums import LanguageModelType, InferenceServiceType
from edsl.Base import RichPrintingMixin, PersistenceMixin
from edsl.data.Cache import Cache
from edsl.enums import service_to_api_keyname

from edsl.exceptions import MissingAPIKeyError


def handle_key_error(func):
    """Handle KeyError exceptions."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
            assert True == False
        except KeyError as e:
            return f"""KeyError occurred: {e}. This is most likely because the model you are using 
            returned a JSON object we were not expecting."""

    return wrapper


class RegisterLanguageModelsMeta(ABCMeta):
    """Metaclass to register output elements in a registry i.e., those that have a parent."""

    _registry = {}  # Initialize the registry as a dictionary
    REQUIRED_CLASS_ATTRIBUTES = ["_model_", "_parameters_", "_inference_service_"]

    def __init__(cls, name, bases, dct):
        """Register the class in the registry if it has a _model_ attribute."""
        super(RegisterLanguageModelsMeta, cls).__init__(name, bases, dct)
        # if name != "LanguageModel":
        if (model_name := getattr(cls, "_model_", None)) is not None:
            RegisterLanguageModelsMeta.check_required_class_variables(
                cls, RegisterLanguageModelsMeta.REQUIRED_CLASS_ATTRIBUTES
            )

            ## Check that model name is valid
            if not LanguageModelType.is_value_valid(model_name):
                acceptable_values = [item.value for item in LanguageModelType]
                raise LanguageModelAttributeTypeError(
                    f"""A LanguageModel's model must be one of {LanguageModelType} values, which are
                    {acceptable_values}. You passed {model_name}."""
                )

            if not InferenceServiceType.is_value_valid(
                inference_service := getattr(cls, "_inference_service_", None)
            ):
                acceptable_values = [item.value for item in InferenceServiceType]
                raise LanguageModelAttributeTypeError(
                    f"""A LanguageModel's model must have an _inference_service_ value from 
                    {acceptable_values}. You passed {inference_service}."""
                )

            # LanguageModel children have to implement the async_execute_model_call method
            RegisterLanguageModelsMeta.verify_method(
                candidate_class=cls,
                method_name="async_execute_model_call",
                expected_return_type=dict[str, Any],
                required_parameters=[("user_prompt", str), ("system_prompt", str)],
                must_be_async=True,
            )
            # LanguageModel children have to implement the parse_response method
            RegisterLanguageModelsMeta.verify_method(
                candidate_class=cls,
                method_name="parse_response",
                expected_return_type=str,
                required_parameters=[("raw_response", dict[str, Any])],
                must_be_async=False,
            )
            RegisterLanguageModelsMeta._registry[model_name] = cls

    @classmethod
    def get_registered_classes(cls):
        """Return the registry."""
        return cls._registry

    @staticmethod
    def check_required_class_variables(
        candidate_class: LanguageModel, required_attributes: List[str] = None
    ):
        """Check if a class has the required attributes.

        >>> class M:
        ...     _model_ = "m"
        ...     _parameters_ = {}
        >>> RegisterLanguageModelsMeta.check_required_class_variables(M, ["_model_", "_parameters_"])
        >>> class M2:
        ...     _model_ = "m"
        >>> RegisterLanguageModelsMeta.check_required_class_variables(M2, ["_model_", "_parameters_"])
        Traceback (most recent call last):
        ...
        Exception: Class M2 does not have required attribute _parameters_
        """
        required_attributes = required_attributes or []
        for attribute in required_attributes:
            if not hasattr(candidate_class, attribute):
                raise Exception(
                    f"Class {candidate_class.__name__} does not have required attribute {attribute}"
                )

    @staticmethod
    def verify_method(
        candidate_class: LanguageModel,
        method_name: str,
        expected_return_type: Any,
        required_parameters: List[tuple[str, Any]] = None,
        must_be_async: bool = False,
    ):
        """Verify that a method is defined in a class, has the correct return type, and has the correct parameters."""
        RegisterLanguageModelsMeta._check_method_defined(candidate_class, method_name)

        required_parameters = required_parameters or []
        method = getattr(candidate_class, method_name)
        signature = inspect.signature(method)

        RegisterLanguageModelsMeta._check_return_type(method, expected_return_type)

        if must_be_async:
            RegisterLanguageModelsMeta._check_is_coroutine(method)

        # Check the parameters
        params = signature.parameters
        for param_name, param_type in required_parameters:
            RegisterLanguageModelsMeta._verify_parameter(
                params, param_name, param_type, method_name
            )

    @staticmethod
    def _check_method_defined(cls, method_name):
        """Check if a method is defined in a class.

        Example:
        >>> class M:
        ...     def f(self): pass
        >>> RegisterLanguageModelsMeta._check_method_defined(M, "f")
        >>> RegisterLanguageModelsMeta._check_method_defined(M, "g")
        Traceback (most recent call last):
        ...
        NotImplementedError: g method must be implemented.
        """
        if not hasattr(cls, method_name):
            raise NotImplementedError(f"{method_name} method must be implemented.")

    @staticmethod
    def _check_is_coroutine(func: Callable):
        """Check to make sure it's a coroutine function.

        Example:

        >>> def f(): pass
        >>> RegisterLanguageModelsMeta._check_is_coroutine(f)
        Traceback (most recent call last):
        ...
        TypeError: A LangugeModel class with method f must be an asynchronous method.
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError(
                f"A LangugeModel class with method {func.__name__} must be an asynchronous method."
            )

    @staticmethod
    def _verify_parameter(params, param_name, param_type, method_name):
        """Verify that a parameter is defined in a method and has the correct type."""
        if param_name not in params:
            raise TypeError(
                f"""Parameter "{param_name}" of method "{method_name}" must be defined.
                """
            )
        if params[param_name].annotation != param_type:
            raise TypeError(
                f"""Parameter "{param_name}" of method "{method_name}" must be of type {param_type.__name__}.
                Got {params[param_name].annotation} instead.
                """
            )

    @staticmethod
    def _check_return_type(method, expected_return_type):
        """
        Check if the return type of a method is as expected.

        Example:
        >>> class M:
        ...     async def f(self) -> str: pass
        >>> RegisterLanguageModelsMeta._check_return_type(M.f, str)
        >>> class N:
        ...     async def f(self) -> int: pass
        >>> RegisterLanguageModelsMeta._check_return_type(N.f, str)
        Traceback (most recent call last):
        ...
        TypeError: Return type of f must be <class 'str'>. Got <class 'int'>.
        """
        if inspect.isroutine(method):
            # return_type = inspect.signature(method).return_annotation
            return_type = get_type_hints(method)["return"]
            if return_type != expected_return_type:
                raise TypeError(
                    f"Return type of {method.__name__} must be {expected_return_type}. Got {return_type}."
                )

    @classmethod
    def model_names_to_classes(cls):
        """Return a dictionary of model names to classes."""
        d = {}
        for classname, cls in cls._registry.items():
            if hasattr(cls, "_model_"):
                d[cls._model_] = cls
            else:
                raise Exception(
                    f"Class {classname} does not have a _model_ class attribute."
                )
        return d


class LanguageModel(
    RichPrintingMixin, PersistenceMixin, ABC, metaclass=RegisterLanguageModelsMeta
):
    """ABC for LLM subclasses."""

    _model_ = None

    __rate_limits = None
    # TODO: Use the OpenAI Teir 1 rate limits
    __default_rate_limits = {"rpm": 10_000, "tpm": 2_000_000}
    _safety_factor = 0.8

    def __init__(self, **kwargs):
        """Initialize the LanguageModel."""
        self.model = getattr(self, "_model_", None)
        default_parameters = getattr(self, "_parameters_", None)
        parameters = self._overide_default_parameters(kwargs, default_parameters)
        self.parameters = parameters

        for key, value in parameters.items():
            setattr(self, key, value)

        for key, value in kwargs.items():
            if key not in parameters:
                setattr(self, key, value)

        if "use_cache" in kwargs:
            warnings.warn(
                "The use_cache parameter is deprecated. Use the Cache class instead."
            )
 
        if skip_api_key_check := kwargs.get("skip_api_key_check", False):
            # Skip the API key check. Sometimes this is useful for testing.
            self.api_token = None

        if not hasattr(self, "api_token"):
            key_name = service_to_api_keyname.get(self._inference_service_, "NOT FOUND")
            self.api_token = os.getenv(key_name)
            if self.api_token is None and self._inference_service_ != 'test':
                raise MissingAPIKeyError(
                    f"""The key for service: `{self._inference_service_}` is not set.
                    Need a key with name {key_name} in your .env file.
                    """
                )

    def __getitem__(self, key):
        return getattr(self, key)

    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html
        return data_to_html(self.to_dict())

    def hello(self, verbose = True):
        """Runs a simple test to check if the model is working."""
        token = self.api_token
        masked = token[:min(8, len(token))] + "..."
        if verbose:
            print(f"Current key is {masked}")
        return self.execute_model_call(user_prompt = "Hello, model!", system_prompt = "You are a helpful agent.")

    def has_valid_api_key(self) -> bool:
        """Check if the model has a valid API key.

        This method is used to check if the model has a valid API key.
        """
        from edsl.enums import service_to_api_keyname
        import os

        if self._model_ == LanguageModelType.TEST.value:
            return True

        key_name = service_to_api_keyname.get(self._inference_service_, "NOT FOUND")
        key_value = os.getenv(key_name)
        return key_value is not None

    def __hash__(self):
        """Allow the model to be used as a key in a dictionary."""
        return hash(self.model + str(self.parameters))

    def __eq__(self, other):
        """Check is two models are the same.

        >>> m1 = LanguageModel.example()
        >>> m2 = LanguageModel.example()
        >>> m1 == m2
        True

        """
        return self.model == other.model and self.parameters == other.parameters

    def _set_rate_limits(self, rpm=None, tpm=None) -> None:
        """Set the rate limits for the model.

        If the model does not have rate limits, use the default rate limits."""
        if rpm is not None and tpm is not None:
            self.__rate_limits = {"rpm": rpm, "tpm": tpm}
            return

        if self.__rate_limits is None:
            if hasattr(self, "get_rate_limits"):
                self.__rate_limits = self.get_rate_limits()
            else:
                self.__rate_limits = self.__default_rate_limits

    @property
    def RPM(self):
        """Model's requests-per-minute limit."""
        self._set_rate_limits()
        return self._safety_factor * self.__rate_limits["rpm"]

    @property
    def TPM(self):
        """Model's tokens-per-minute limit."""
        self._set_rate_limits()
        return self._safety_factor * self.__rate_limits["tpm"]

    @staticmethod
    def _overide_default_parameters(passed_parameter_dict, default_parameter_dict):
        """Return a dictionary of parameters, with passed parameters taking precedence over defaults.

        >>> LanguageModel._overide_default_parameters(passed_parameter_dict={"temperature": 0.5}, default_parameter_dict={"temperature":0.9})
        {'temperature': 0.5}
        >>> LanguageModel._overide_default_parameters(passed_parameter_dict={"temperature": 0.5}, default_parameter_dict={"temperature":0.9, "max_tokens": 1000})
        {'temperature': 0.5, 'max_tokens': 1000}
        """
        parameters = dict({})
        for parameter, default_value in default_parameter_dict.items():
            if parameter in passed_parameter_dict:
                parameters[parameter] = passed_parameter_dict[parameter]
            else:
                parameters[parameter] = default_value
        return parameters

    @abstractmethod
    async def async_execute_model_call():
        """Execute the model call and returns the result as a coroutine."""
        pass

    @jupyter_nb_handler
    def execute_model_call(self, *args, **kwargs) -> Coroutine:
        """Execute the model call and returns the result as a coroutine."""

        async def main():
            results = await asyncio.gather(
                self.async_execute_model_call(*args, **kwargs)
            )
            return results[0]  # Since there's only one task, return its result

        return main()

    @abstractmethod
    def parse_response(raw_response: dict[str, Any]) -> str:
        """Parse the API response and returns the response text.

        What is returned by the API is model-specific and often includes meta-data that we do not need.
        For example, here is the results from a call to GPT-4:
        To actually tract the response, we need to grab
        data["choices[0]"]["message"]["content"].
        """
        raise NotImplementedError

    def _update_response_with_tracking(
        self, response, start_time, cached_response=False, cache_key=None
    ):
        """Update the response with tracking information and post it to the API Queue."""
        end_time = time.time()
        response["elapsed_time"] = end_time - start_time
        response["timestamp"] = end_time
        response["cached_response"] = cached_response
        response["cache_key"] = cache_key
        return response

    async def async_get_raw_response(
        self,
        user_prompt: str,
        system_prompt: str,
        cache,
        iteration: int = 0,
    ) -> dict[str, Any]:
        """Handle caching of responses.

        :param user_prompt: The user's prompt.
        :param system_prompt: The system's prompt.
        :param iteration: The iteration number.
        :param cache: The cache to use.

        If the cache isn't being used, it just returns a 'fresh' call to the LLM,
        but appends some tracking information to the response (using the _update_response_with_tracking method).
        But if cache is being used, it first checks the database to see if the response is already there.
        If it is, it returns the cached response, but again appends some tracking information.
        If it isn't, it calls the LLM, saves the response to the database, and returns the response with tracking information.

        If self.use_cache is True, then attempts to retrieve the response from the database;
        if not in the DB, calls the LLM and writes the response to the DB.
        """
        start_time = time.time()

        cached_response = cache.fetch(
            model=str(self.model),
            parameters=self.parameters,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            iteration=iteration,
        )

        if cache_used := (cached_response is not None):
            # print("Cache used")
            response = json.loads(cached_response)
        else:
            # print("Cache not used")
            # print(f"Cache data is: {cache.data}")
            response = await self.async_execute_model_call(user_prompt, system_prompt)

        if not cache_used:
            cache_key = cache.store(
                user_prompt=user_prompt,
                model=str(self.model),
                parameters=self.parameters,
                system_prompt=system_prompt,
                response=response,
                iteration=iteration,
            )
        else:
            cache_key = None
        return self._update_response_with_tracking(
            response=response,
            start_time=start_time,
            cached_response=cache_used,
            cache_key=cache_key,
        )

    get_raw_response = sync_wrapper(async_get_raw_response)

    def simple_ask(self, question: QuestionBase, system_prompt = "You are a helpful agent pretending to be a human.", top_logprobs = 2):
        """Ask a question and return the response."""
        self.logprobs = True
        self.top_logprobs = top_logprobs
        return self.execute_model_call(user_prompt=question.human_readable(), system_prompt=system_prompt)

    async def async_get_response(
        self, user_prompt: str, system_prompt: str, cache: Cache, iteration: int = 1
    ) -> dict:
        """Get response, parse, and return as string.

        :param user_prompt: The user's prompt.
        :param system_prompt: The system's prompt.
        :param iteration: The iteration number.
        :param cache: The cache to use.

        """
        raw_response = await self.async_get_raw_response(
            user_prompt=user_prompt,
            system_prompt=system_prompt,
            iteration=iteration,
            cache=cache,
        )
        response = self.parse_response(raw_response)
        try:
            dict_response = json.loads(response)
        except json.JSONDecodeError as e:
            # TODO: Turn into logs to generate issues
            dict_response, success = await repair(response, str(e))
            if not success:
                raise Exception("Even the repair failed.")

        dict_response["cached_response"] = raw_response["cached_response"]
        dict_response["usage"] = raw_response.get("usage", {})
        dict_response["raw_model_response"] = raw_response
        return dict_response

    get_response = sync_wrapper(async_get_response)

    def cost(self, raw_response: dict[str, Any]) -> float:
        """Return the dollar cost of a raw response."""
        keys = raw_response["usage"].keys()
        prices = model_prices.get(self.model)
        return sum([prices.get(key, 0.0) * raw_response["usage"][key] for key in keys])

    #######################
    # SERIALIZATION METHODS
    #######################
    def to_dict(self) -> dict[str, Any]:
        """Convert instance to a dictionary."""
        return {"model": self.model, "parameters": self.parameters}

    @classmethod
    def from_dict(cls, data: dict) -> Type[LanguageModel]:
        """Convert dictionary to a LanguageModel child instance."""
        from edsl.language_models.registry import get_model_class

        model_class = get_model_class(data["model"])
        data["use_cache"] = True
        return model_class(**data)

    #######################
    # DUNDER METHODS
    #######################
    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"{self.__class__.__name__}(model = '{self.model}', parameters={self.parameters})"

    def __add__(self, other_model: Type[LanguageModel]) -> Type[LanguageModel]:
        """Combine two models into a single model (other_model takes precedence over self)."""
        print(
            f"""Warning: one model is replacing another. If you want to run both models, use a single `by` e.g., 
              by(m1, m2, m3) not by(m1).by(m2).by(m3)."""
        )
        return other_model or self

    def rich_print(self):
        """Display an object as a table."""
        table = Table(title="Language Model")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))

        return table

    @classmethod
    def example(cls):
        """Return a default instance of the class."""
        from edsl import Model

        return Model(Model.available()[0], skip_api_key_check=True)


if __name__ == "__main__":
    """Run the module's test suite."""
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    from edsl.language_models import LanguageModel

    print(LanguageModel.example())



from __future__ import annotations
import random
import textwrap
from typing import Any, Optional, Union
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import IntegerDescriptor, QuestionOptionsDescriptor
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionBudget(QuestionBase):
    """This question prompts the agent to allocate a budget among options."""

    question_type = "budget"
    budget_sum: int = IntegerDescriptor(none_allowed=False)
    question_options: list[str] = QuestionOptionsDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        budget_sum: int,
        short_names_dict: Optional[dict[str, str]] = None,
    ):
        """Instantiate a new QuestionBudget.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param question_options: The options for allocation of the budget sum.
        :param budget_sum: The total amount of the budget to be allocated among the options.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionBudget.default_instructions`.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.question_options = question_options
        self.budget_sum = budget_sum
        self.short_names_dict = short_names_dict or {}

    ################
    # Answer methods
    ################
    def _validate_answer(self, answer: dict[str, Any]) -> dict[str, Union[int, str]]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_key_value(answer, "answer", dict)
        self._validate_answer_budget(answer)
        return answer

    def _translate_answer_code_to_answer(
        self, answer_codes: dict[str, int], scenario: Scenario = None
    ):
        """
        Translate the answer codes to the actual answers.

        For example, for a budget question with options ["a", "b", "c"],
        the answer codes are 0, 1, and 2. The LLM will respond with 0.
        This code will translate that to "a".
        """
        translated_codes = []
        for answer_code, response in answer_codes.items():
            translated_codes.append({self.question_options[int(answer_code)]: response})

        return translated_codes

    def _simulate_answer(self, human_readable=True):
        """Simulate a valid answer for debugging purposes (what the validator expects)."""
        if human_readable:
            keys = self.question_options
        else:
            keys = range(len(self.question_options))
        remaining_budget = self.budget_sum
        values = []
        for _ in range(len(self.question_options)):
            if _ == len(self.question_options) - 1:
                # Assign remaining budget to the last value
                values.append(remaining_budget)
            else:
                # Generate a random value between 0 and remaining budget
                value = random.randint(0, remaining_budget)
                values.append(value)
                remaining_budget -= value
        answer = dict(zip(keys, values))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionBudget:
        """Return an example of a budget question."""
        return cls(
            question_name="food_budget",
            question_text="How would you allocate $100?",
            question_options=["Pizza", "Ice Cream", "Burgers", "Salad"],
            budget_sum=100,
        )


def main():
    """Create an example of a budget question and demonstrate its functionality."""
    from edsl.questions.QuestionBudget import QuestionBudget

    q = QuestionBudget.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    # validate an answer
    q._validate_answer(
        {"answer": {0: 100, 1: 0, 2: 0, 3: 0}, "comment": "I like custard"}
    )
    # translate answer code
    q._translate_answer_code_to_answer({0: 100, 1: 0, 2: 0, 3: 0})
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q


if __name__ == "__main__":
    q = QuestionBudget.example()
    results = q.run()

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.enums import LanguageModelType
from edsl.language_models.DeepInfra import create_deep_infra_model

model_name = LanguageModelType.MIXTRAL_8x7B_INSTRUCT.value
url = "https://api.deepinfra.com/v1/inference/mistralai/Mixtral-8x7B-Instruct-v0.1"

Mixtral8x7B = create_deep_infra_model(
    model_name=model_name, url=url, model_class_name="Mixtral8x7B"
)

if __name__ == "__main__":
    from edsl.questions import QuestionMultipleChoice

    m = Mixtral8x7B()
    q = QuestionMultipleChoice(
        question_text="Are pickled pigs feet a popular breakfast food in the US?",
        question_options=["Yes", "No", "Unsure"],
        question_name="bkfast_question",
    )
    results = q.by(m).run()
    from rich import print

    print(q)
    results.select("answer.*", "model.model").print()



# import anthropic
import os
import dotenv
from typing import Any
from anthropic import AsyncAnthropic

dotenv.load_dotenv()

api_key = os.environ.get("ANTHROPIC_API_KEY")

from edsl.language_models.LanguageModel import LanguageModel
from edsl.enums import InferenceServiceType
from edsl.enums import LanguageModelType
from edsl.exceptions import MissingAPIKeyError
import os
import re


def create_anthropic_model(model_name, model_class_name) -> LanguageModel:
    class LLM(LanguageModel):
        """
        Child class of LanguageModel for interacting with OpenAI models
        """

        _inference_service_ = InferenceServiceType.ANTHROPIC.value
        _model_ = model_name
        _parameters_ = {
            "temperature": 0.5,
            "max_tokens": 1000,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "logprobs": False,
            "top_logprobs": 3,
        }

        async def async_execute_model_call(
            self, user_prompt: str, system_prompt: str = ""
        ) -> dict[str, Any]:
            """Calls the OpenAI API and returns the API response."""
            api_key = os.environ.get("ANTHROPIC_API_KEY")
            client = AsyncAnthropic(api_key=api_key)

            response = await client.messages.create(
                model="claude-3-opus-20240229",
                max_tokens=1000,
                temperature=0.0,
                system=system_prompt,
                messages=[
                    #     {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
            )
            return response.model_dump()

        @staticmethod
        def parse_response(raw_response: dict[str, Any]) -> str:
            """Parses the API response and returns the response text."""
            response = raw_response["content"][0]["text"]
            pattern = r"^```json(?:\\n|\n)(.+?)(?:\\n|\n)```$"
            match = re.match(pattern, response, re.DOTALL)
            if match:
                return match.group(1)
            else:
                return response

    LLM.__name__ = model_class_name

    return LLM


if __name__ == "__main__":
    pass
    # ClaudeOpus = create_anthropic_model("claude-3-opus-20240229", "ClaudeOpus")
    # results = m.execute_model_call("How are you today?")
    # cleaned_up = ClaudeOpus.parse_response(results)
    # print(cleaned_up)



from collections import UserDict


class TaskStatusLogEntry(UserDict):
    def __init__(self, log_time, value):
        self.data = {"log_time": log_time, "value": value}
        super().__init__(self.data)



"""Free text question type."""
import textwrap
from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class FreeText(QuestionInstuctionsBase):
    """Free text question type."""

    question_type = "free_text"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        Return a valid JSON formatted like this:
        {"answer": "<put free text answer here>"}
        """
    )



from __future__ import annotations
import json
import os
import warnings
from typing import Optional, Union
from edsl.config import CONFIG
from edsl.data.CacheEntry import CacheEntry
from edsl.data.SQLiteDict import SQLiteDict


EDSL_DATABASE_PATH = CONFIG.get("EDSL_DATABASE_PATH")
EXPECTED_PARROT_CACHE_URL = os.getenv("EXPECTED_PARROT_CACHE_URL")

# TODO: What do we want to do if & when there is a mismatch
#       -- if two keys are the same but the values are different?
# TODO: In the read methods, if the file already exists, make sure it is valid.


class Cache:
    """
    A class that represents a cache of responses from a language model.

    :param data: The data to initialize the cache with.
    :param remote: Whether to sync the Cache with the server.
    :param immediate_write: Whether to write to the cache immediately after storing a new entry.

    Deprecated:

    :param method: The method of storage to use for the cache.
    """

    data = {}

    def __init__(
        self,
        *,
        data: Optional[Union[SQLiteDict, dict]] = None,
        remote: bool = False,
        immediate_write: bool = True,
        method=None,
    ):
        """
        Create two dictionaries to store the cache data.

        :param new_entries: Entries that are created during a __enter__ block.
        :param new_entries_to_write_later: Entries that will be written to the cache later.
        """
        self.data = data or {}
        self.remote = remote
        self.immediate_write = immediate_write
        self.method = method
        self.new_entries = {}
        self.new_entries_to_write_later = {}
        self.coop = None
        self._perform_checks()

    def keys(self):
        return list(self.data.keys())
    
    def values(self):
        return list(self.data.values())

    def _perform_checks(self):
        """Perform checks on the cache."""
        if any(not isinstance(value, CacheEntry) for value in self.data.values()):
            raise Exception("Not all values are CacheEntry instances")
        if self.method is not None:
            warnings.warn("Argument `method` is deprecated", DeprecationWarning)
        if self.remote:
            from edsl.coop import Coop

            self.coop = Coop()

    ####################
    # READ/WRITE
    ####################
    def fetch(
        self,
        *,
        model: str,
        parameters: dict,
        system_prompt: str,
        user_prompt: str,
        iteration: int,
    ) -> Union[None, str]:
        """
        Fetch a value (LLM output) from the cache.
        Return None if the response is not found.
        """
        key = CacheEntry.gen_key(
            model=model,
            parameters=parameters,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            iteration=iteration,
        )
        entry = self.data.get(key, None)
        return None if entry is None else entry.output

    def store(
        self,
        model: str,
        parameters: str,
        system_prompt: str,
        user_prompt: str,
        response: str,
        iteration: int,
    ) -> str:
        """
        Add a new key-value pair to the cache.

        * Key is a hash of the input parameters.
        * Output is the response from the language model.

        How it works:

        * The key-value pair is added to `self.new_entries`
        * If `immediate_write` is True , the key-value pair is added to `self.data`
        * If `immediate_write` is False, the key-value pair is added to `self.new_entries_to_write_later`
        """
        entry = CacheEntry(
            model=model,
            parameters=parameters,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            output=json.dumps(response),
            iteration=iteration,
        )
        key = entry.key
        self.new_entries[key] = entry
        if self.immediate_write:
            self.data[key] = entry
        else:
            self.new_entries_to_write_later[key] = entry
        return key

    def add_from_dict(
        self, new_data: dict[str, CacheEntry], write_now: Optional[bool] = True
    ) -> None:
        """
        Add entries to the cache from a dictionary.

        :param write_now: Whether to write to the cache immediately (similar to `immediate_write`).
        """
        for key, value in new_data.items():
            if key in self.data:
                if value != self.data[key]:
                    raise Exception("Mismatch in values")
            if not isinstance(value, CacheEntry):
                raise Exception(f"Wrong type - the observed type is {type(value)}")

        self.new_entries.update(new_data)
        if write_now:
            self.data.update(new_data)
        else:
            self.new_entries_to_write_later.update(new_data)

    def add_from_jsonl(self, filename: str, write_now: Optional[bool] = True) -> None:
        """
        Add entries to the cache from a JSONL.

        :param write_now: Whether to write to the cache immediately (similar to `immediate_write`).
        """
        with open(filename, "a+") as f:
            f.seek(0)
            lines = f.readlines()
        new_data = {}
        for line in lines:
            d = json.loads(line)
            key = list(d.keys())[0]
            value = list(d.values())[0]
            new_data[key] = CacheEntry(**value)
        self.add_from_dict(new_data=new_data, write_now=write_now)

    def add_from_sqlite(self, db_path: str, write_now: Optional[bool] = True):
        """
        Add entries to the cache from an SQLite database.

        :param write_now: Whether to write to the cache immediately (similar to `immediate_write`).
        """
        db = SQLiteDict(db_path)
        new_data = {}
        for key, value in db.items():
            new_data[key] = CacheEntry(**value)
        self.add_from_dict(new_data=new_data, write_now=write_now)

    @classmethod
    def from_sqlite_db(cls, db_path: str) -> Cache:
        """
        Construct a Cache from a SQLite database.
        """
        return cls(data=SQLiteDict(db_path))

    @classmethod
    def from_jsonl(cls, jsonlfile: str, db_path: str = None) -> Cache:
        """
        Construct a Cache from a JSONL file.

        * If `db_path` is None, the cache will be stored in memory, as a dictionary.
        * If `db_path` is provided, the cache will be stored in an SQLite database.
        """
        if db_path is None:
            data = {}
        else:
            data = SQLiteDict(db_path)
        
        # if a file doesn't exist at jsonfile, throw an error
        if not os.path.exists(jsonlfile):
            raise FileNotFoundError(f"File {jsonlfile} not found")

        cache = Cache(data=data)
        cache.add_from_jsonl(jsonlfile)
        return cache

    ## TODO: Check to make sure not over-writing (?)
    ## Should be added to SQLiteDict constructor (?)
    def write_sqlite_db(self, db_path: str) -> None:
        """
        Write the cache to an SQLite database.
        """
        new_data = SQLiteDict(db_path)
        for key, value in self.data.items():
            new_data[key] = value

    def write_jsonl(self, filename: str) -> None:
        """
        Write the cache to a JSONL file.
        """
        path = os.path.join(os.getcwd(), filename)
        with open(path, "w") as f:
            for key, value in self.data.items():
                f.write(json.dumps({key: value.to_dict()}) + "\n")

    ####################
    # REMOTE
    ####################
    # TODO: Make this work
    # - Need to decide whether the cache belongs to a user and what can be shared
    # - I.e., some cache entries? all or nothing?
    @classmethod
    def from_url(cls, db_path=None) -> Cache:
        """
        Construct a Cache object from a remote.
        """
        # ...do something here
        # return Cache(data=db)
        pass

    def __enter__(self):
        """
        Run when a context is entered.
        """
        if self.remote:
            print("Syncing local and remote caches")
            exclude_keys = list(self.data.keys())
            cache_entries = self.coop.get_cache_entries(exclude_keys)
            self.add_from_dict({c.key: c for c in cache_entries}, write_now=True)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Run when a context is exited.
        """
        for key, entry in self.new_entries_to_write_later.items():
            self.data[key] = entry
        if self.remote:
            _ = self.coop.send_cache_entries(cache_dict=self.new_entries)

    ####################
    # DUNDER / USEFUL
    ####################
    def to_dict(self) -> dict:
        """Return the Cache as a dictionary."""
        return {k: v.to_dict() for k, v in self.data.items()}
    
    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html
        return data_to_html(self.to_dict())

    @classmethod
    def from_dict(cls, data) -> Cache:
        """Construct a Cache from a dictionary."""
        data = {k: CacheEntry.from_dict(v) for k, v in data}
        return cls(data=data)

    def __len__(self):
        """Return the number of CacheEntry objects in the Cache."""
        return len(self.data)

    # TODO: Same inputs could give different results and this could be useful
    # can't distinguish unless we do the ε trick or vary iterations
    def __eq__(self, other_cache: "Cache") -> bool:
        """
        Check if two Cache objects are equal.
        Does not verify their values are equal, only that they have the same keys.
        """
        if not isinstance(other_cache, Cache):
            return False
        return set(self.data.keys()) == set(other_cache.data.keys())

    def __add__(self, other: "Cache"):
        """
        Combine two caches.
        """
        if not isinstance(other, Cache):
            raise ValueError("Can only add two caches together")
        self.data.update(other.data)
        return self

    def __repr__(self):
        """
        Return a string representation of the Cache object.
        """
        return f"Cache(data = {repr(self.data)}, immediate_write={self.immediate_write}, remote={self.remote})"

    ####################
    # EXAMPLES
    ####################
    def fetch_input_example(self) -> dict:
        """
        Create an example input for a 'fetch' operation.
        """
        return CacheEntry.fetch_input_example()

    def to_html(self):
        #json_str = json.dumps(self.data, indent=4)
        d = {k: v.to_dict() for k, v in self.data.items()}
        for key, value in d.items():
            for k, v in value.items():
                if isinstance(v, dict):
                    d[key][k] = {kk: str(vv) for kk, vv in v.items()}
                else:
                    d[key][k] = str(v) 

        json_str = json.dumps(d, indent=4)

        # HTML template with the JSON string embedded
        html = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>Display JSON</title>
        </head>
        <body>
            <pre id="jsonData"></pre>
            <script>
                var json = {json_str};

                // JSON.stringify with spacing to format
                document.getElementById('jsonData').textContent = JSON.stringify(json, null, 4);
            </script>
        </body>
        </html>
        """
        return html
    
    def view(self) -> None:
        """View the Cache in a new browser tab."""
        import tempfile
        import webbrowser
        html_content = self.to_html()
        # Create a temporary file to hold the HTML
        with tempfile.NamedTemporaryFile('w', delete=False, suffix='.html') as tmpfile:
            tmpfile.write(html_content)
            # Get the path to the temporary file
            filepath = tmpfile.name

        # Open the HTML file in a new browser tab
        webbrowser.open('file://' + filepath)


    @classmethod
    def example(cls) -> Cache:
        """
        Return an example Cache.
        The example Cache has one entry.
        """
        return cls(data={CacheEntry.example().key: CacheEntry.example()})


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""This file contains the configuration for the prompt generation."""
from enum import Enum

NEGATIVE_INFINITY = float("-inf")


class AttributeTypes(Enum):
    """The types of attributes that a prompt can have."""

    COMPONENT_TYPE = "component_type"
    MODEL = "model"
    QUESTION_TYPE = "question_type"


class ComponentTypes(Enum):
    """The types of attributes that a prompt can have."""

    TEST = "test"
    GENERIC = "generic"
    QUESTION_DATA = "question_data"
    QUESTION_INSTRUCTIONS = "question_instructions"
    AGENT_INSTRUCTIONS = "agent_instructions"
    AGENT_PERSONA = "agent_persona"
    SURVEY_INSTRUCTIONS = "survey_instructions"
    SURVEY_DATA = "survey_data"


names_to_component_types = {v.value: v for k, v in ComponentTypes.__members__.items()}

C2A = {
    ComponentTypes.QUESTION_INSTRUCTIONS: [
        AttributeTypes.QUESTION_TYPE,
        AttributeTypes.MODEL,
    ],
    ComponentTypes.AGENT_INSTRUCTIONS: [AttributeTypes.MODEL],
}



import textwrap

DEFAULT_MODEL_CLASS = "edsl.language_models.LanguageModelOpenAIFour"

from edsl.language_models.LanguageModel import RegisterLanguageModelsMeta
import pkgutil
import importlib
import edsl.language_models.model_interfaces as model_interfaces

# Dynamically import all modules in the model_interfaces package
for loader, module_name, is_pkg in pkgutil.iter_modules(model_interfaces.__path__):
    full_module_name = f"{model_interfaces.__name__}.{module_name}"
    if not is_pkg:
        module = importlib.import_module(full_module_name)
        globals().update(
            {
                name: getattr(module, name)
                for name in dir(module)
                if not name.startswith("_")
            }
        )

meta_class_registry = RegisterLanguageModelsMeta.get_registered_classes()

# For compatibility with older versions of EDSL
get_model_class = (
    lambda model_name: RegisterLanguageModelsMeta.model_names_to_classes().get(
        model_name
    )
)


class Meta(type):
    def __repr__(cls):
        return textwrap.dedent(
            f"""\
        Available models: {cls.available()}
        
        To create an instance, you can do: 
        >>> m = Model('gpt-4-1106-preview', temperature=0.5, ...)
        
        To get the default model, you can leave out the model name. 
        To see the available models, you can do:
        >>> Model.available()
        """
        )


class Model(metaclass=Meta):
    default_model = "gpt-4-1106-preview"

    def __new__(cls, model_name=None, *args, **kwargs):
        # Map index to the respective subclass
        get_model_classes = RegisterLanguageModelsMeta.model_names_to_classes()

        if cls.default_model not in get_model_classes:
            raise ValueError(f"Default model {cls.default_model} not found")

        if model_name is None:
            model_name = cls.default_model
            #print(f"No model name provided, using default model: {model_name}")
        
        subclass = get_model_classes.get(model_name, None)
        if subclass is None:
            raise ValueError(
                f"No model registered with name '{model_name}'. See a list of available models run: Model.available()"
            )

        # Create an instance of the selected subclass
        instance = object.__new__(subclass)
        instance.__init__(*args, **kwargs)
        return instance

    @classmethod
    def available(cls):
        return sorted(list(RegisterLanguageModelsMeta.model_names_to_classes().keys()))
    
    @classmethod
    def check_models(cls, verbose = False):
        print("Checking all available models...\n")
        for model in cls.available():
            print(f"Now checking: {model}")
            try:
                m = cls(model)
            except Exception as e:
                print(f"Error creating instance of {model}: {e}")
                continue 
            try:
                results = m.hello(verbose)
                if verbose:
                    print(f"Results from model call: {results}")
            except Exception as e:
                print(f"Error calling 'hello' on {model}: {e}")
                continue
            print("OK!")
            print("\n")


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    available = Model.available()
    m = Model("gpt-4-1106-preview")
    results = m.execute_model_call("Hello world")
    print(results)



"""A Survey is collection of questions that can be administered to an Agent."""
from __future__ import annotations
import re
from rich import print
from rich.table import Table

from dataclasses import dataclass

from typing import Any, Generator, Optional, Union, List
from edsl.exceptions import SurveyCreationError, SurveyHasNoRulesError
from edsl.questions.QuestionBase import QuestionBase
from edsl.surveys.base import RulePriority, EndOfSurvey
from edsl.surveys.Rule import Rule
from edsl.surveys.RuleCollection import RuleCollection

from edsl.Base import Base
from edsl.surveys.SurveyExportMixin import SurveyExportMixin
from edsl.surveys.descriptors import QuestionsDescriptor
from edsl.surveys.MemoryPlan import MemoryPlan

from edsl.surveys.DAG import DAG

from edsl.utilities import is_notebook




@dataclass
class SurveyMetaData:
    """Metadata for a survey. This is a dataclass that holds the name, description, and version of a survey."""

    name: str = None
    description: str = None
    version: str = None


from edsl.surveys.SurveyFlowVisualizationMixin import SurveyFlowVisualizationMixin


class Survey(SurveyExportMixin, SurveyFlowVisualizationMixin, Base):
    """A collection of questions that supports skip logic."""

    questions = QuestionsDescriptor()
    """
    A collection of questions that supports skip logic.

    Initalization:
    - `questions`: the questions in the survey (optional)
    - `question_names`: the names of the questions (optional)
    - `name`: the name of the survey (optional)

    Methods:
    -

    Notes:
    - The presumed order of the survey is the order in which questions are added.
    """

    def __init__(
        self,
        questions: list[QuestionBase] = None,
        memory_plan: MemoryPlan = None,
        name: str = None,
        description: str = None,
        version: str = None,
    ):
        """Create a new survey."""
        self.rule_collection = RuleCollection(
            num_questions=len(questions) if questions else None
        )
        self.meta_data = SurveyMetaData(
            name=name, description=description, version=version
        )
        self.questions = questions or []
        self.memory_plan = memory_plan or MemoryPlan(self)

    @property
    def name(self):
        """Return the name of the survey."""
        # print("WARNING: name is deprecated. Please use meta_data.name instead.")
        return self.meta_data.name

    def get_question(self, question_name) -> QuestionBase:
        """Return the question object given the question name."""
        if question_name not in self.question_name_to_index:
            raise KeyError(f"Question name {question_name} not found in survey.")
        index = self.question_name_to_index[question_name]
        return self._questions[index]

    @property
    def question_names(self) -> list[str]:
        """Return a list of question names in the survey.

        Example:
        >>> s = Survey.example()
        >>> s.question_names
        ['q0', 'q1', 'q2']
        """
        # return list(self.question_name_to_index.keys())
        return [q.question_name for q in self.questions]

    @property
    def question_name_to_index(self) -> dict[str, int]:
        """Return a dictionary mapping question names to question indices.

        Example:
        >>> s = Survey.example()
        >>> s.question_name_to_index
        {'q0': 0, 'q1': 1, 'q2': 2}
        """
        return {q.question_name: i for i, q in enumerate(self.questions)}

    def add_question(
        self, question: QuestionBase, question_name: Optional[str] = None
    ) -> Survey:
        """
        Add a question to survey.

        - The question is appended at the end of the self.questions list
        - A default rule is created that the next index is the next question.
        """
        if question_name is not None:
            print(
                "Warning: question_name is deprecated. Please use question.question_name instead."
            )

        if question.question_name in self.question_names:
            raise SurveyCreationError(
                f"""Question name already exists in survey. Please use a different name for the offensing question.
                The problemetic question name is {question_name}.
                """
            )
        index = len(self.questions)
        # TODO: This is a bit ugly because the user
        # doesn't "know" about _questions - it's generated by the
        # descriptor.
        self._questions.append(question)

        # using index + 1 presumes there is a next question
        self.rule_collection.add_rule(
            Rule(
                current_q=index,
                expression="True",
                next_q=index + 1,
                question_name_to_index=self.question_name_to_index,
                priority=RulePriority.DEFAULT.value,
            )
        )

        # a question might be added before the memory plan is created
        # it's ok because the memory plan will be updated when it is created
        if hasattr(self, "memory_plan"):
            self.memory_plan.add_question(question)

        return self

    def set_full_memory_mode(self) -> None:
        """Add instructions to a survey that the agent should remember all of the answers to the questions in the survey."""
        self._set_memory_plan(lambda i: self.question_names[:i])

    def set_lagged_memory(self, lags: int):
        """Add instructions to a survey that the agent should remember the answers to the questions in the survey.

        The agent should remember the answers to the questions in the survey from the previous lags.
        """
        self._set_memory_plan(lambda i: self.question_names[max(0, i - lags) : i])

    def _set_memory_plan(self, prior_questions_func):
        """Set memory plan based on a provided function determining prior questions."""
        for i, question_name in enumerate(self.question_names):
            self.memory_plan.add_memory_collection(
                focal_question=question_name,
                prior_questions=prior_questions_func(i),
            )

    def add_targeted_memory(
        self,
        focal_question: Union[QuestionBase, str],
        prior_question: Union[QuestionBase, str],
    ) -> Survey:
        """Add instructions to a survey than when answering focal_question.

        The agent should also remember the answers to prior_questions listed in prior_questions.
        """
        focal_question_name = self.question_names[
            self._get_question_index(focal_question)
        ]
        prior_question_name = self.question_names[
            self._get_question_index(prior_question)
        ]

        self.memory_plan.add_single_memory(
            focal_question=focal_question_name,
            prior_question=prior_question_name,
        )

        return self

    def add_memory_collection(
        self,
        focal_question: Union[QuestionBase, str],
        prior_questions: List[Union[QuestionBase, str]],
    ):
        """Add prior questions and responses so the agent has them when answering.

        This adds instructions to a survey than when answering focal_question, the agent should also remember the answers to prior_questions listed in prior_questions.

        :param focal_question: The question that the agent is answering.
        :param prior_questions: The questions that the agent should remember when answering the focal question.

        Example:
        
        >>> s = Survey.example()
        >>> s.add_memory_collection("q2", ["q0", "q1"])
        """
        focal_question_name = self.question_names[
            self._get_question_index(focal_question)
        ]

        prior_question_names = [
            self.question_names[self._get_question_index(prior_question)]
            for prior_question in prior_questions
        ]

        self.memory_plan.add_memory_collection(
            focal_question=focal_question_name, prior_questions=prior_question_names
        )

    def add_stop_rule(self, question: Union[QuestionBase, str], expression: str) -> Survey:
        """Add a rule that stops the survey."""
        self.add_rule(question, expression, EndOfSurvey)
        return self

    def _get_question_index(self, q:Union[QuestionBase, str, EndOfSurvey.__class__]) -> int:
        """Return the index of the question or EndOfSurvey object.

        It can handle it if the user passes in the question name, the question object, or the EndOfSurvey object.
        """
        if q == EndOfSurvey:
            return EndOfSurvey
        else:
            question_name = q if isinstance(q, str) else q.question_name
            if question_name not in self.question_name_to_index:
                raise ValueError(
                    f"""Question name {question_name} not found in survey. The current question names are {self.question_name_to_index}."""
                )
            return self.question_name_to_index[question_name]

    def add_rule(
        self, question: QuestionBase, expression: str, next_question: QuestionBase
    ) -> Survey:
        """
        Add a rule to a Question of the Survey with the appropriate priority.

        - The last rule added for the question will have the highest priority.
        - If there are no rules, the rule added gets priority -1.
        """
        question_index = self._get_question_index(question)
        next_question_index = self._get_question_index(next_question)

        def get_new_rule_priority(question_index):
            """Return the priority for the new rule."""
            current_priorities = [
                rule.priority
                for rule in self.rule_collection.applicable_rules(question_index)
            ]
            max_priority = max(current_priorities)
            # newer rules take priority over older rules
            new_priority = (
                RulePriority.DEFAULT.value
                if len(current_priorities) == 0
                else max_priority + 1
            )
            return new_priority

        self.rule_collection.add_rule(
            Rule(
                current_q=question_index,
                expression=expression,
                next_q=next_question_index,
                question_name_to_index=self.question_name_to_index,
                priority=get_new_rule_priority(question_index),
            )
        )

        return self

    ###################
    # FORWARD METHODS
    ###################
    def by(self, *args: Union[Agent, Scenario, LanguageModel]) -> Jobs:
        """Add Agents, Scenarios, and LanguageModels to a survey and returns a runnable Jobs object."""
        from edsl.jobs.Jobs import Jobs

        job = Jobs(survey=self)
        return job.by(*args)

    def run(self, *args, **kwargs) -> Jobs:
        """Turn the survey into a Job and runs it."""
        from edsl.jobs.Jobs import Jobs

        return Jobs(survey=self).run(*args, **kwargs)

    ########################
    ## Survey-Taking Methods
    ########################

    def first_question(self) -> QuestionBase:
        """Return the first question in the survey."""
        return self.questions[0]

    def next_question(
        self, current_question: "Question", answers: dict
    ) -> Union[QuestionBase, EndOfSurvey.__class__]:
        """
        Return the next question in a survey.

        - If called with no arguments, it returns the first question in the survey.
        - If no answers are provided for a question with a rule, the next question is returned. If answers are provided, the next question is determined by the rules and the answers.
        - If the next question is the last question in the survey, an EndOfSurvey object is returned.
        """
        if isinstance(current_question, str):
            print(
                "WARNING: current_question by string is deprecated. Please use a Question object."
            )
            current_question = self.get_question(current_question)

        question_index = self.question_name_to_index[current_question.question_name]
        next_question_object = self.rule_collection.next_question(
            question_index, answers
        )

        if next_question_object.num_rules_found == 0:
            raise SurveyHasNoRulesError

        if next_question_object.next_q == EndOfSurvey:
            return EndOfSurvey
        else:
            if next_question_object.next_q >= len(self.questions):
                return EndOfSurvey
            else:
                return self.questions[next_question_object.next_q]

    def gen_path_through_survey(self) -> Generator[QuestionBase, dict, None]:
        """
        Generate a coroutine that can be used to conduct an Interview.

        - The coroutine is a generator that yields a question and receives answers.
        - The coroutine starts with the first question in the survey.
        - The coroutine ends when an EndOfSurvey object is returned.

        E.g., in Interview.py

        path_through_survey = self.survey.gen_path_through_survey()
        question = path_through_survey.send({question.question_name: answer})
        """
        question = self.first_question()
        while not question == EndOfSurvey:
            self.answers = yield question
            question = self.next_question(question, self.answers)

    @property
    def scenario_attributes(self) -> list[str]:
        """Return a list of attributes that admissible Scenarios should have."""
        temp = []
        for question in self.questions:
            question_text = question.question_text
            # extract the contents of all {{ }} in the question text using regex
            matches = re.findall(r"\{\{(.+?)\}\}", question_text)
            # remove whitespace
            matches = [match.strip() for match in matches]
            # add them to the temp list
            temp.extend(matches)
        return temp

    def textify(self, index_dag: DAG) -> DAG:
        """Convert the DAG of question indices to a DAG of question names.

        Example:
        >>> s = Survey.example()
        >>> d = s.dag()
        >>> d
        {1: {0}, 2: {0}}
        >>> s.textify(d)
        {'q1': {'q0'}, 'q2': {'q0'}}
        """

        def get_name(index):
            """Return the name of the question given the index."""
            return self.questions[index].question_name

        try:
            text_dag = {}
            for child_index, parent_indices in index_dag.items():
                parent_names = {get_name(index) for index in parent_indices}
                child_name = get_name(child_index)
                text_dag[child_name] = parent_names
            return text_dag
        except IndexError:
            breakpoint()

    def dag(self, textify=False) -> DAG:
        """Return the DAG of the survey, which reflects both skip-logic and memory."""
        memory_dag = self.memory_plan.dag
        rule_dag = self.rule_collection.dag
        if textify:
            memory_dag = DAG(self.textify(memory_dag))
            rule_dag = DAG(self.textify(rule_dag))
        return memory_dag + rule_dag

    ###################
    # DUNDER METHODS
    ###################
    def __len__(self) -> int:
        """Return the number of questions in the survey."""
        return len(self._questions)

    def __getitem__(self, index) -> Question:
        """Return the question object given the question index."""
        if isinstance(index, int):
            return self._questions[index]
        elif isinstance(index, str):
            return getattr(self, index)

    def __eq__(self, other):
        """Return True if the two surveys have the same to_dict."""
        if not isinstance(other, Survey):
            return False
        return self.to_dict() == other.to_dict()

    ###################
    # SERIALIZATION METHODS
    ###################
    def to_dict(self) -> dict[str, Any]:
        """Serialize the Survey object to a dictionary."""
        return {
            "questions": [q.to_dict() for q in self._questions],
            "name": self.name,
            "memory_plan": self.memory_plan.to_dict(),
            "rule_collection": self.rule_collection.to_dict(),
        }

    @classmethod
    def from_dict(cls, data: dict) -> Survey:
        """Deserialize the dictionary back to a Survey object."""
        questions = [QuestionBase.from_dict(q_dict) for q_dict in data["questions"]]
        memory_plan = MemoryPlan.from_dict(data["memory_plan"])
        survey = cls(questions=questions, name=data["name"], memory_plan=memory_plan)
        survey.rule_collection = RuleCollection.from_dict(data["rule_collection"])
        return survey

    ###################
    # DISPLAY METHODS
    ###################
    def __repr__(self) -> str:
        """Return a string representation of the survey."""
        questions_string = ", ".join([repr(q) for q in self._questions])
        question_names_string = ", ".join([repr(name) for name in self.question_names])
        return f"Survey(questions=[{questions_string}], name={repr(self.name)})"
        
    def _repr_html_(self) -> str:
        from edsl.utilities.utilities import data_to_html
        return data_to_html(self.to_dict())
  
    def show_rules(self) -> None:
        """Print out the rules in the survey."""
        self.rule_collection.show_rules()

    def rich_print(self):
        """Print the survey in a rich format."""
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Questions", style="dim")

        for question in self._questions:
            table.add_row(question.rich_print())

        return table

    def show_questions(self):
        """Print out the questions in the survey."""
        for name, question in zip(self.question_names, self._questions):
            print(f"Question: {name}")
            print(question)

    def codebook(self) -> dict[str, str]:
        """Create a codebook for the survey, mapping question names to question text."""
        codebook = {}
        for question in self._questions:
            codebook[question.question_name] = question.question_text
        return codebook

    @classmethod
    def example(cls) -> Survey:
        """Return an example survey."""
        from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice

        # from edsl.surveys.Survey import Survey

        q0 = QuestionMultipleChoice(
            question_text="Do you like school?",
            question_options=["yes", "no"],
            question_name="q0",
        )
        q1 = QuestionMultipleChoice(
            question_text="Why not?",
            question_options=["killer bees in cafeteria", "other"],
            question_name="q1",
        )
        q2 = QuestionMultipleChoice(
            question_text="Why?",
            question_options=["**lack*** of killer bees in cafeteria", "other"],
            question_name="q2",
        )
        s = cls(questions=[q0, q1, q2])
        s = s.add_rule(q0, "q0 == 'yes'", q2)
        return s


def main():
    """Run the example survey."""

    def example_survey():
        """Return an example survey."""
        from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        from edsl.surveys.Survey import Survey

        q0 = QuestionMultipleChoice(
            question_text="Do you like school?",
            question_options=["yes", "no"],
            question_name="q0",
        )
        q1 = QuestionMultipleChoice(
            question_text="Why not?",
            question_options=["killer bees in cafeteria", "other"],
            question_name="q1",
        )
        q2 = QuestionMultipleChoice(
            question_text="Why?",
            question_options=["**lack*** of killer bees in cafeteria", "other"],
            question_name="q2",
        )
        s = Survey(questions=[q0, q1, q2])
        s = s.add_rule(q0, "q0 == 'yes'", q2)
        return s

    s = example_survey()
    survey_dict = s.to_dict()
    s2 = Survey.from_dict(survey_dict)
    results = s2.run()
    print(results)


if __name__ == "__main__":

    def example_survey():
        """Return an example survey."""
        from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        from edsl.surveys.Survey import Survey

        q0 = QuestionMultipleChoice(
            question_text="Do you like school?",
            question_options=["yes", "no"],
            question_name="like_school",
        )
        q1 = QuestionMultipleChoice(
            question_text="Why not?",
            question_options=["killer bees in cafeteria", "other"],
            question_name="why_not",
        )
        q2 = QuestionMultipleChoice(
            question_text="Why?",
            question_options=["**lack*** of killer bees in cafeteria", "other"],
            question_name="why",
        )
        s = Survey(questions=[q0, q1, q2])
        s = s.add_rule(q0, "like_school == 'yes'", q2).add_stop_rule(
            q1, "why_not == 'other'"
        )
        return s

    # s = example_survey()
    # survey_dict = s.to_dict()
    # s2 = Survey.from_dict(survey_dict)
    # results = s2.run()
    # print(results)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    s = example_survey()
    s.show_flow()



"""A module for status logs and status objects.

A StatusObject is a dictionary. 

For example, for a given task, it could be:
{
    'status': TaskStatus.SUCCESS,
    'tokens_used': 100,
}

Then, an aggregate like an InterviewStatusObject is some aggregation of these status objects.

These are then stored in a log as entries with a time stamp.
The aggregation, again at the level of a specific task, could be 
{
    'final_status': TaskStatus.SUCCESS,
    'elapsed_time': 100,
}

For an interview, it could be: 
{
    'status': InterviewStatus.WAITING_FOR_RESOURCES,
    'tokens_used': 10_00,
}
"""
from abc import ABC, abstractmethod
import time

from collections import UserList, UserDict


class StatusObjectABC(ABC, UserDict):
    """An abstract base class for status objects.

    This is a snapshot of this object at a moment in time.
    -For example, for a task, it could be just the state of the task.
    -For an interivew, could be the states of the all the component tasks.
    -For a job, it could be the states of all the interviews.

    An abstract base class for status objects.
    """

    @abstractmethod
    def state(self):
        """Return the state of the status object, which must be from a specified enum."""
        pass

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the status object."""
        pass

    @abstractmethod
    def to_dict(self) -> dict:
        """Return the status object as a dictionary."""
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, data_dict) -> "StatusObjectABC":
        """Return a status object from a dictionary."""
        pass


class CumulativeStatusObjectABC(ABC):
    """An abstract base class for cumulative status objects."""

    def __init__(self, value):
        """Initialize the object with a value."""
        self.value = value

    def add(self, new_object):
        """Add a new object to the cumulative object."""
        pass


class StatusLogABC(ABC, UserList):
    """For example, these could be status of a task, or the status of a job."""

    def __init__(self, data=None):
        """Initialize the log with the given data."""
        if data is None:
            data = []
        super().__init__(data)
        self._cumulative_state = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Check that the subclass has a status_enum attribute."""
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "status_enum"):
            raise TypeError(f"{cls.__name__} is missing 'status_enum' class attribute")

    @property
    def current(self):
        """Return the most recently added status."""
        self._check()
        return self[-1]

    def _check(self):
        """Check that the log is not empty."""
        if len(self) == 0:
            raise Exception("No data in log")

    # def reduce(self):
    #     """Reduce a list of status objects to a single status object."""
    #     _, result = self[0]
    #     for _, status in self[1:]:
    #         result += status
    #     return result

    def log(self, status) -> None:
        """Check to make sure time is later."""
        t = time.monotonic()
        if len(self) > 0 and self[-1][0] > t:
            raise ValueError("Time must be later than previous time")

        self.append((t, status))

    def status_at_time(self, target_time):
        """Return the status at a given time."""
        for t, status in reversed(self.data):
            if t <= target_time:
                return status
        return None

    @abstractmethod
    def __repr__(self):
        """Return a string representation of the log."""
        pass

    @property
    def min_time(self):
        """Return the earliest time in the log."""
        if len(self) == 0:
            raise Exception("No data in log")
        return self[0][0]

    @property
    def max_time(self):
        """Return the latest time in the log."""
        if len(self) == 0:
            raise Exception("No data in log")
        return self[-1][0]

    def to_dict(self):
        """Return the log as a dictionary."""
        return {"data": [(t, status.to_dict()) for t, status in self.data]}

    @classmethod
    def from_dict(cls, data_dict):
        """Return a log from a dictionary."""
        obj = cls()
        data = data_dict["data"]
        for t, status in data:
            obj.append(t, status.from_dict())

    def status_vector(self, num_periods=10):
        """Return a matrix of status values."""
        start_time = self.min_time
        end_time = self.max_time
        time_increment = (end_time - start_time) / num_periods
        time_periods = [start_time + i * time_increment for i in range(num_periods)]
        status_vector = [self.status_at_time(t) for t in time_periods]
        return status_vector

    def state_vector(self, num_periods=10):
        """Return a vector of status values."""
        status_vector = self.status_vector(num_periods)
        return [status.state() for status in status_vector]

    def state_vector_with_colors(self, num_periods):
        """Return a vector of status values."""
        status_vector = self.status_vector(num_periods)
        return [self.status_color_dict[status] for status in status_vector]

    # def numerical_matrix(self, num_periods):
    #     """Return a numerical matrix of status values."""
    #     status_dicts = self.status_vector(num_periods)

    #     num_cols = num_periods
    #     num_rows = len(status_dicts)
    #     matrix = [[0 for _ in range(num_cols)] for _ in range(num_rows)]

    #     for row_index, (task_name, status_list) in enumerate(status_dicts.items()):
    #         matrix[row_index] = [list(status_colors.keys()).index(status) for status in status_list]

    #     index_to_names = {i: name for i, name in enumerate(status_dicts.keys())}
    #     return matrix, index_to_names

    # def visualize(self, num_periods=10):
    #     """Visualize the status matrix with outlined squares."""
    #     import matplotlib.pyplot as plt
    #     from matplotlib.colors import ListedColormap
    #     import numpy as np
    #     from matplotlib.patches import Rectangle

    #     # Define your custom colormap
    #     custom_cmap = ListedColormap(list(status_colors.values()))

    #     # Generate the matrix
    #     matrix, index_to_names = self.numerical_matrix(num_periods)

    #     # Create the figure and axes
    #     plt.figure(figsize=(10, 5))
    #     ax = plt.gca()

    #     # Display the matrix and keep a reference to the imshow object
    #     im = ax.imshow(matrix, aspect='auto', cmap=custom_cmap)

    #     # Adding color bar, now correctly associating it with 'im'
    #     cbar = plt.colorbar(im, ticks=range(len(status_colors)), label='Task Status')

    #     cbar_labels = [status.name for status in status_colors.keys()]
    #     #breakpoint()
    #     cbar.set_ticklabels(cbar_labels)  # Setting the custom labels for the colorbar

    #     im.set_clim(-0.5, len(status_colors) - 0.5)  # Setting color limits directly on the imshow object

    #     # Outline each cell by drawing rectangles
    #     for (j, i), val in np.ndenumerate(matrix):
    #         ax.add_patch(Rectangle((i - 0.5, j - 0.5), 1, 1, fill=False, edgecolor='black', lw=0.5))

    #     # Set custom y-axis ticks and labels
    #     yticks = list(index_to_names.keys())
    #     yticklabels = list(index_to_names.values())
    #     plt.yticks(ticks=yticks, labels=yticklabels)

    #     # Show the plot
    #     plt.show()


if __name__ == "__main__":
    import doctest

    doctest.testmod()

    import enum

    class ExampleStatusEnum(enum.Enum):
        """An example of a status enum."""

        NOT_STARTED = enum.auto()
        IN_PROGRESS = enum.auto()
        SUCCESS = enum.auto()
        FAILED = enum.auto()

    class ExampleStatus(StatusObjectABC):
        """An example of a status object."""

        status_enum = ExampleStatusEnum
        status_color_dict = {
            ExampleStatusEnum.NOT_STARTED: "grey",
            ExampleStatusEnum.IN_PROGRESS: "orange",
            ExampleStatusEnum.SUCCESS: "green",
            ExampleStatusEnum.FAILED: "red",
        }

        def __init__(self, value):
            """Initialize the status object with a value."""
            self.value = value

        def state(self):
            """Return the state of the status object."""
            return self.value

        def __repr__(self):
            """Return a string representation of the status object."""
            return f"ExampleStatus(value = {self.value})"

        def to_dict(self):
            """Return the status object as a dictionary."""
            return {"example_key": "example_value"}

        @classmethod
        def from_dict(cls, data_dict):
            """Return a status object from a dictionary."""
            return cls()

    class ExampleStatusLog(StatusLogABC):
        """An example of a status log."""

        status_enum = ExampleStatusEnum

        def __repr__(self):
            """Return a string representation of the log."""
            return f"ExampleStatusLog(data = {[(t, repr(status)) for t, status in self.data]})"

    example_status1 = ExampleStatus(ExampleStatusEnum.NOT_STARTED)
    example_status2 = ExampleStatus(ExampleStatusEnum.IN_PROGRESS)
    example_status3 = ExampleStatus(ExampleStatusEnum.SUCCESS)

    log = ExampleStatusLog()
    log.log(example_status1)
    log.log(example_status2)
    log.log(example_status3)

    print(log.to_dict())

    log.current
    log.min_time
    log.max_time
    log.status_vector()



import os
import json
import requests
import threading

class ReportErrors:

    def __init__(self, task_history):
        self.task_history = task_history
        self.email = None
        self.url = os.getenv("EXPECTED_PARROT_ERROR_REPORTING_URL", None)
        if self.url is None:
            raise ValueError("The URL for the error reporting service is not set.")

    @property
    def data(self):
        return {
            "text": self.task_history.to_dict(),
            "email": self.email,
        }

    def get_email(self, timeout=10):
        """ Helper method to get user input with a timeout. """
        input_queue = []

        def input_thread_method():
            email_input = input("Please enter your email address (if you want us to get in touch): ")
            input_queue.append(email_input)

        input_thread = threading.Thread(target=input_thread_method)
        input_thread.start()
        input_thread.join(timeout=timeout)

        if input_queue:
            self.email = input_queue[0]
        else:
            print("No input received within the timeout period.")

    def upload(self):
        json_data = json.dumps(self.data)
        headers = {
            "Content-Type": "application/json"
        }
        response = requests.post(self.url, data=json_data, headers=headers)
        print("Status Code:", response.status_code)




import time
import asyncio
import textwrap
from typing import Coroutine, List, AsyncGenerator

from rich.live import Live
from rich.console import Console

from edsl.results import Results, Result

# from edsl.jobs.runners.JobsRunner import JobsRunner
from edsl.jobs.interviews.Interview import Interview
from edsl.utilities.decorators import jupyter_nb_handler
from edsl.jobs.Jobs import Jobs
from edsl.utilities.utilities import is_notebook
from edsl.jobs.runners.JobsRunnerStatusMixin import JobsRunnerStatusMixin

# from edsl.jobs.runners.JobsRunHistory import JobsRunHistory

from edsl.data.Cache import Cache

# from edsl.jobs.tasks.task_status_enum import TaskStatus

from edsl.jobs.tasks.TaskHistory import TaskHistory


class JobsRunnerAsyncio(JobsRunnerStatusMixin):
    def __init__(self, jobs: Jobs):
        self.jobs = jobs

        self.interviews: List["Interview"] = jobs.interviews()
        self.bucket_collection: "BucketCollection" = jobs.bucket_collection
        self.total_interviews: List["Interview"] = []

    def populate_total_interviews(self, n=1) -> None:
        """Populates self.total_interviews with n copies of each interview.

        :param n: how many times to run each interview.
        """
        self.total_interviews = []
        for interview in self.interviews:
            for iteration in range(n):
                if iteration > 0:
                    new_interview = Interview(
                        agent=interview.agent,
                        survey=interview.survey,
                        scenario=interview.scenario,
                        model=interview.model,
                        debug=interview.debug,
                        iteration=iteration,
                        cache=self.cache,
                    )
                    self.total_interviews.append(new_interview)
                else:
                    interview.cache = self.cache
                    self.total_interviews.append(interview)

    async def run_async(
        self,
        cache,
        n: int = 1,
        debug: bool = False,
        stop_on_exception: bool = False,
        sidecar_model = None,
    ) -> AsyncGenerator[Result, None]:
        """Creates the tasks, runs them asynchronously, and returns the results as a Results object.

        Completed tasks are yielded as they are completed.

        :param n: how many times to run each interview
        :param debug:
        :param stop_on_exception:
        """
        tasks = []
        self.populate_total_interviews(
            n=n
        )  # Populate self.total_interviews before creating tasks

        for interview in self.total_interviews:
            interviewing_task = self._interview_task(
                interview=interview, debug=debug, stop_on_exception=stop_on_exception, sidecar_model=sidecar_model
            )
            tasks.append(asyncio.create_task(interviewing_task))

        for task in asyncio.as_completed(tasks):
            result = await task
            yield result

    async def _interview_task(
        self, *, interview: Interview, debug: bool, stop_on_exception: bool = False, sidecar_model = None
    ) -> Result:
        """Conducts an interview and returns the result.

        :param interview: the interview to conduct
        :param debug: prints debug messages
        """
        # the model buckets are used to track usage rates
        model_buckets = self.bucket_collection[interview.model]

        # get the results of the interview
        answer, valid_results = await interview.async_conduct_interview(
            debug=debug,
            model_buckets=model_buckets,
            stop_on_exception=stop_on_exception,
            sidecar_model=sidecar_model,
        )

        # we should have a valid result for each question
        answer_key_names = {k for k in set(answer.keys()) if not k.endswith("_comment")}

        assert len(valid_results) == len(answer_key_names)

        # TODO: move this down into Interview
        question_name_to_prompts = dict({})
        for result in valid_results:
            question_name = result["question_name"]
            question_name_to_prompts[question_name] = {
                "user_prompt": result["prompts"]["user_prompt"],
                "system_prompt": result["prompts"]["system_prompt"],
            }

        prompt_dictionary = {}
        for answer_key_name in answer_key_names:
            prompt_dictionary[
                answer_key_name + "_user_prompt"
            ] = question_name_to_prompts[answer_key_name]["user_prompt"]
            prompt_dictionary[
                answer_key_name + "_system_prompt"
            ] = question_name_to_prompts[answer_key_name]["system_prompt"]

        raw_model_results_dictionary = {}
        for result in valid_results:
            question_name = result["question_name"]
            raw_model_results_dictionary[
                question_name + "_raw_model_response"
            ] = result["raw_model_response"]

        result = Result(
            agent=interview.agent,
            scenario=interview.scenario,
            model=interview.model,
            iteration=interview.iteration,
            answer=answer,
            prompt=prompt_dictionary,
            raw_model_response=raw_model_results_dictionary,
            survey=interview.survey,
        )
        return result

    @property
    def elapsed_time(self):
        return time.monotonic() - self.start_time

    @jupyter_nb_handler
    async def run(
        self,
        cache,
        n: int = 1,
        debug: bool = False,
        stop_on_exception: bool = False,
        progress_bar=False,
        sidecar_model=None,
        batch_mode = False
    ) -> "Coroutine":
        """Runs a collection of interviews, handling both async and sync contexts."""
        console = Console()
        self.results = []
        self.start_time = time.monotonic()
        self.completed = False
        self.cache = cache
        self.sidecar_model = sidecar_model

        def generate_table():
            return self.status_table(self.results, self.elapsed_time)

        from contextlib import contextmanager

        @contextmanager
        def no_op_cm():
            """A no-op context manager with a dummy update method."""
            yield DummyLive()

        class DummyLive:
            def update(self, *args, **kwargs):
                """A dummy update method that does nothing."""
                pass

        progress_bar_context = (
            Live(generate_table(), console=console, refresh_per_second=5)
            if progress_bar
            else no_op_cm()
        )

        with cache as c:
            with progress_bar_context as live:

                async def update_progress_bar():
                    """Updates the progress bar at fixed intervals."""
                    while True:
                        live.update(generate_table())
                        await asyncio.sleep(0.1)  # Update interval
                        if self.completed:
                            break

                async def process_results():
                    """Processes results from interviews."""
                    async for result in self.run_async(
                        n=n, debug=debug, stop_on_exception=stop_on_exception, cache=c, 
                        sidecar_model=sidecar_model
                    ):
                        self.results.append(result)
                        live.update(generate_table())
                    self.completed = True

                # logger_task = asyncio.create_task(self.periodic_logger(period=0.01))
                progress_task = asyncio.create_task(update_progress_bar())

                try:
                    await asyncio.gather(process_results(), progress_task)
                except asyncio.CancelledError:
                    pass
                finally:
                    progress_task.cancel()  # Cancel the progress_task when process_results is done
                    await progress_task

                    await asyncio.sleep(1)  # short delay to show the final status

                    # one more update
                    live.update(generate_table())

        results = Results(survey=self.jobs.survey, data=self.results)
        results.task_history = TaskHistory(self.total_interviews, include_traceback=False)

        if results.task_history.has_exceptions and not batch_mode:
            print(
                textwrap.dedent(f"""\Exceptions were raised in the following interviews: {results.task_history.indices}.
                The object results.task_history contains the exceptions.                
                """
                )
            )
            show = input("Print exceptions? (y/n): ")
            if show == "y":
                if is_notebook():
                    from IPython.display import HTML, display
                    display(HTML(results.task_history._repr_html_()))
                else:
                    results.task_history.show_exceptions()
                try:
                    from edsl.jobs.interviews.ReportErrors import ReportErrors
                    full_task_history = TaskHistory(self.total_interviews, include_traceback=True)
                    report = ReportErrors(full_task_history)
                    upload = input("Ok to upload errors to us? We can potentially help! (y/n): ")
                    if upload == "y":
                        report.get_email()
                        report.upload()
                        print("Errors are reported here: ", report.url)
                except Exception as e:
                    pass 
        return results



"""This module contains the Interview class, which is responsible for conducting an interview asynchronously."""

from __future__ import annotations
import asyncio
import time
import traceback
from typing import Generator
from edsl import CONFIG
from edsl.exceptions import InterviewTimeoutError
from edsl.data_transfer_models import AgentResponseDict
from edsl.questions.QuestionBase import QuestionBase
from edsl.surveys.base import EndOfSurvey
from edsl.jobs.buckets.ModelBuckets import ModelBuckets
from edsl.jobs.interviews.interview_exception_tracking import InterviewExceptionEntry
from edsl.jobs.interviews.retry_management import retry_strategy
from edsl.jobs.tasks.task_status_enum import TaskStatus
from edsl.jobs.tasks.TasksList import TasksList
from edsl.jobs.tasks.QuestionTaskCreator import QuestionTaskCreator
from edsl.data.Cache import Cache

TIMEOUT = float(CONFIG.get("EDSL_API_TIMEOUT"))


class InterviewTaskBuildingMixin:
    def _build_invigilators(self, debug: bool) -> Generator["Invigilator", None, None]:
        """Create an invigilator for each question."""
        for question in self.survey.questions:
            yield self.get_invigilator(question=question, debug=debug)

    def get_invigilator(self, question: QuestionBase, debug: bool) -> "Invigilator":
        """Return an invigilator for the given question."""
        invigilator = self.agent.create_invigilator(
            question=question,
            scenario=self.scenario,
            model=self.model,
            debug=debug,
            memory_plan=self.survey.memory_plan,
            current_answers=self.answers,
            iteration=self.iteration,
            cache=self.cache,
            sidecar_model=self.sidecar_model,
        )
        """Return an invigilator for the given question."""
        return invigilator

    @property
    def dag(self) -> "DAG":
        """Return the directed acyclic graph for the survey.

        The DAG, or directed acyclic graph, is a dictionary that maps question names to their dependencies.
        It is used to determine the order in which questions should be answered.
        This reflects both agent 'memory' considerations and 'skip' logic.
        The 'textify' parameter is set to True, so that the question names are returned as strings rather than integer indices.
        """
        return self.survey.dag(textify=True)

    def _build_question_tasks(
        self,
        debug: bool,
        model_buckets: ModelBuckets,
    ) -> list[asyncio.Task]:
        """Create a task for each question, with dependencies on the questions that must be answered before this one can be answered."""
        tasks = []
        for question in self.survey.questions:
            tasks_that_must_be_completed_before = list(
                self._get_tasks_that_must_be_completed_before(
                    tasks=tasks, question=question
                )
            )
            question_task = self._create_question_task(
                question=question,
                tasks_that_must_be_completed_before=tasks_that_must_be_completed_before,
                model_buckets=model_buckets,
                debug=debug,
                iteration=self.iteration,
            )
            tasks.append(question_task)
        return TasksList(tasks)  # , invigilators

    def _get_tasks_that_must_be_completed_before(
        self, *, tasks: list[asyncio.Task], question: QuestionBase
    ) -> Generator[asyncio.Task, None, None]:
        """Return the tasks that must be completed before the given question can be answered.

        If a question has no dependencies, this will be an empty list, [].
        """
        parents_of_focal_question = self.dag.get(question.question_name, [])
        for parent_question_name in parents_of_focal_question:
            parent_index = self.to_index[parent_question_name]
            parent_task = tasks[parent_index]
            yield parent_task

    def _create_question_task(
        self,
        *,
        question: QuestionBase,
        tasks_that_must_be_completed_before: list[asyncio.Task],
        model_buckets: ModelBuckets,
        debug: bool,
        iteration: int = 0,
    ) -> asyncio.Task:
        """Create a task that depends on the passed-in dependencies that are awaited before the task is run.

        The task is created by a QuestionTaskCreator, which is responsible for creating the task and managing its dependencies.
        It is passed a reference to the function that will be called to answer the question.
        It is passed a list "tasks_that_must_be_completed_before" that are awaited before the task is run.
        These are added as a dependency to the focal task.
        """
        task_creator = QuestionTaskCreator(
            question=question,
            answer_question_func=self._answer_question_and_record_task,
            token_estimator=self._get_estimated_request_tokens,
            model_buckets=model_buckets,
            iteration=iteration,
        )
        for task in tasks_that_must_be_completed_before:
            task_creator.add_dependency(task)

        self.task_creators.update(
            {question.question_name: task_creator}
        )  # track this task creator
        return task_creator.generate_task(debug)

    def _get_estimated_request_tokens(self, question) -> float:
        """Estimate the number of tokens that will be required to run the focal task."""
        invigilator = self.get_invigilator(question=question, debug=False)
        # TODO: There should be a way to get a more accurate estimate.
        combined_text = ""
        for prompt in invigilator.get_prompts().values():
            if hasattr(prompt, "text"):
                combined_text += prompt.text
            elif isinstance(prompt, str):
                combined_text += prompt
            else:
                raise ValueError(f"Prompt is of type {type(prompt)}")
        return len(combined_text) / 4.0

    @retry_strategy
    async def _answer_question_and_record_task(
        self,
        *,
        question: QuestionBase,
        debug: bool,
        task=None,
    ) -> AgentResponseDict:
        """Answer a question and records the task.

        This in turn calls the the passed-in agent's async_answer_question method, which returns a response dictionary.
        Note that is updates answers with the response.
        """
        invigilator = self.get_invigilator(question, debug=debug)

        async def attempt_to_answer_question(invigilator):
            try:
                return await asyncio.wait_for(
                    invigilator.async_answer_question(), timeout=TIMEOUT
                )
            except asyncio.TimeoutError as e:
                exception_entry = InterviewExceptionEntry(
                    exception=repr(e),
                    time=time.time(),
                    traceback=traceback.format_exc(),
                )
                if task:
                    task.task_status = TaskStatus.FAILED
                self.exceptions.add(question.question_name, exception_entry)

                raise InterviewTimeoutError(f"Task timed out after {TIMEOUT} seconds.")
            except Exception as e:
                exception_entry = InterviewExceptionEntry(
                    exception=repr(e),
                    time=time.time(),
                    traceback=traceback.format_exc(),
                )
                if task:
                    task.task_status = TaskStatus.FAILED
                self.exceptions.add(question.question_name, exception_entry)
                raise e

        response: AgentResponseDict = await attempt_to_answer_question(invigilator)

        self.answers.add_answer(response=response, question=question)
        self._cancel_skipped_questions(question)

        return AgentResponseDict(**response)

    def _cancel_skipped_questions(self, current_question: Question) -> None:
        """Cancel the tasks for questions that are skipped.

        It first determines the next question, given the current question and the current answers.
        If the next question is the end of the survey, it cancels all remaining tasks.
        If the next question is after the current question, it cancels all tasks between the current question and the next question.
        """
        current_question_index = self.to_index[current_question.question_name]
        next_question = self.survey.rule_collection.next_question(
            q_now=current_question_index, answers=self.answers
        )
        next_question_index = next_question.next_q

        def cancel_between(start, end):
            """Cancel the tasks between the start and end indices."""
            for i in range(start, end):
                self.tasks[i].cancel()

        if next_question_index == EndOfSurvey:
            cancel_between(current_question_index + 1, len(self.survey.questions))
            return

        if next_question_index > (current_question_index + 1):
            cancel_between(current_question_index + 1, next_question_index)



"""This module contains the RegisterPromptsMeta metaclass, which is used to register prompts."""
import traceback
from collections import defaultdict
from typing import List, Any

from abc import ABCMeta, abstractmethod

from edsl.prompts.prompt_config import (
    C2A,
    names_to_component_types,
    ComponentTypes,
    NEGATIVE_INFINITY,
)

from edsl.enums import QuestionType, LanguageModelType

from edsl.exceptions.prompts import (
    PromptBadQuestionTypeError,
    PromptBadLanguageModelTypeError,
)


class RegisterPromptsMeta(ABCMeta):
    """Metaclass to register prompts."""

    _registry = defaultdict(list)  # Initialize the registry as a dictionary
    _prompts_by_component_type = defaultdict(list)
    # _instances = {}

    # def __new__(mcs, name, bases, dct):
    #     if mcs not in mcs._instances:
    #         mcs._instances[mcs] = super(RegisterPromptsMeta, mcs).__new__(
    #             mcs, name, bases, dct
    #         )
    #     return mcs._instances[mcs]

    def __init__(cls, name, bases, dct):
        """
        We can only have one prompt class per name.

        Each prompt class must have a component type from the ComponentTypes enum.

        Example usage:
        >>> class Prompt1(PromptBase):
        ...     component_type = ComponentTypes.TEST

        >>> class Prompt1(PromptBase):
        ...     component_type = ComponentTypes.TEST
        Traceback (most recent call last):
        ...
        Exception: We already have a Prompt class named Prompt1.
        """
        super(RegisterPromptsMeta, cls).__init__(name, bases, dct)
        # print(f"Current state of registry: {RegisterPromptsMeta._registry}")
        # print(f"Registry called with {name}")
        if "Base" in name or name == "Prompt":
            # print("Exiting")
            return None  # We don't want to register the base class

        if name in RegisterPromptsMeta._registry:
            if RegisterPromptsMeta._registry[name] != cls:
                raise Exception(f"We already have a Prompt class named {name}.")
            else:
                # print("It's the same thing - it's fine.")
                return None

        RegisterPromptsMeta._registry[name] = cls
        # print(f"Current registry: {RegisterPromptsMeta._registry}")
        if (
            component_type := getattr(cls, "component_type", None)
        ) not in ComponentTypes:
            raise Exception(f"Prompt {name} is not in the list of component types")

        ## Make sure that the prompt has a question_type class attribute & it's valid
        if component_type == ComponentTypes.QUESTION_INSTRUCTIONS:
            if not hasattr(cls, "question_type"):
                raise PromptBadQuestionTypeError(
                    "A QuestionInstructions prompt must has a question_type value"
                )
            if not QuestionType.is_value_valid(cls.question_type):
                acceptable_values = [item.value for item in QuestionType]
                raise PromptBadQuestionTypeError(
                    f"""
                A Prompt's question_type must be one of {QuestionType} values, which are 
                currently {acceptable_values}. You passed {cls.question_type}."""
                )

        ## Make sure that if the prompt has a model class attribute, it's valid
        if hasattr(cls, "model"):
            if not LanguageModelType.is_value_valid(cls.model):
                acceptable_values = [item.value for item in LanguageModelType]
                raise PromptBadLanguageModelTypeError(
                    f"""
                A Prompt's model must be one of {LanguageModelType} values, which are 
                currently {acceptable_values}. You passed {cls.model}."""
                )

        key = cls._create_prompt_class_key(dct, component_type)
        cls.data = key
        RegisterPromptsMeta._prompts_by_component_type[component_type].append(cls)

    @classmethod
    def _create_prompt_class_key(cls, dct, component_type) -> tuple[tuple[str, Any]]:
        """Create a key for the prompt class.

        This is a helper function.
        """
        attributes = [attribute.value for attribute in C2A.get(component_type, [])]
        cls_data = {key: value for key, value in dct.items() if key in attributes}
        return tuple(cls_data.items())

    @classmethod
    def _get_classes_with_scores(cls, **kwargs) -> List[tuple[float, "PromptBase"]]:
        """
        Find matching prompts.

        NB that _get_classes_with_scores returns a list of tuples.
        The first element of the tuple is the score, and the second element is the prompt class.
        There is a public-facing function called get_classes that returns only the prompt classes.

        The kwargs are the attributes that we want to match on. E.g., supposed you
        wanted a prompt with component_type = "question_instructions" and question_type = "multiple_choice".
        You would run:

        >>> get_classes(component_type="question_instructions", question_type="multiple_choice", model="gpt-4-1106-preview")
        [<class '__main__.MultipleChoice'>, <class '__main__.MultipleChoiceTurbo'>]

        In the above example, we have two prompts that match. Note that the order of the prompts is determined by the score and the regular MultipleChoice
        is ranked higher because it matches on the model as well.

        Scores are computed by the _score method. The score is the number of attributes that match, with their weights.
        However, if a required attribute doesn't match, then the score is -inf and it can never be selected.

        The function will throw an exception if you don't specify a component type that's in the ComponentTypes enum.

        >>> get_classes(component_type="chicken_tenders", question_type="multiple_choice")
        Traceback (most recent call last):
        ...
        Exception: You must specify a component type. It must be one of dict_keys([...])

        >>> get_classes(component_type="generic")
        []
        """
        component_type_string = kwargs.get("component_type", None)
        component_type = names_to_component_types.get(component_type_string, None)

        if component_type is None:
            raise Exception(
                f"You must specify a component type. It must be one of {names_to_component_types.keys()}"
            )

        try:
            prompts = cls._prompts_by_component_type[component_type]
        except KeyError:
            raise Exception(f"No prompts for component type {component_type}")

        with_scores = [(cls._score(kwargs, prompt), prompt) for prompt in prompts]
        with_scores = sorted(with_scores, key=lambda x: -x[0])
        # filter out the ones with -inf
        matches_with_scores = cls._filter_out_non_matches(with_scores)
        return matches_with_scores

    @classmethod
    def _filter_out_non_matches(cls, prompts_with_scores):
        """Filter out the prompts that have a score of -inf."""
        return [
            (score, prompt)
            for score, prompt in prompts_with_scores
            if score > NEGATIVE_INFINITY
        ]

    @classmethod
    def get_classes(cls, **kwargs):
        """Return only the prompt classes and not the scores.

        Public-facing function.
        """
        with_scores = cls._get_classes_with_scores(**kwargs)
        return [prompt for _, prompt in with_scores]
        # return with_scores

    @classmethod
    def _score(cls, kwargs, prompt):
        """Score the prompt based on the attributes that match."""
        required_list = ["question_type"]
        score = 0
        for key, value in kwargs.items():
            if prompt_value := getattr(prompt, key, None) == value:
                score += 1
            else:
                if key in required_list:
                    score += NEGATIVE_INFINITY
        return score

    @classmethod
    def get_registered_classes(cls):
        """Return the registry."""
        return cls._registry


get_classes = RegisterPromptsMeta.get_classes



"""Base class for all classes in the package. It provides rich printing and persistence of objects."""

from abc import ABC, abstractmethod, ABCMeta
import gzip
import io
import json
from IPython.display import display
from rich.console import Console
from edsl.utilities import is_notebook


class RichPrintingMixin:
    """Mixin for rich printing and persistence of objects."""

    def _for_console(self):
        """Return a string representation of the object for console printing."""
        with io.StringIO() as buf:
            console = Console(file=buf, record=True)
            table = self.rich_print()
            console.print(table)
            return console.export_text()

    def __str__(self):
        """Return a string representation of the object for console printing."""
        return self._for_console()

    def print(self):
        """Print the object to the console."""
        if is_notebook():
            display(self.rich_print())
        else:
            from rich.console import Console

            console = Console()
            console.print(self.rich_print())


class PersistenceMixin:
    """Mixin for saving and loading objects to and from files."""

    def push(self, public=False):
        """Post the object to coop."""
        from edsl.coop import Coop

        c = Coop()
        _ = c.create(self, public)
        print(_)

    @classmethod
    def pull(cls, id):
        """Pull the object from coop."""
        from edsl.coop import Coop

        c = Coop()
        return c._get_base(cls, id)

    @classmethod
    def search(cls, query):
        """Search for objects on coop."""
        from edsl.coop import Coop

        c = Coop()
        return c.search(cls, query)

    def save(self, filename):
        """Save the object to a file as zippped JSON.

        >>> obj.save("obj.json.gz")

        """
        with gzip.open(filename, "wb") as f:
            f.write(json.dumps(self.to_dict()).encode("utf-8"))

    @classmethod
    def load(cls, filename):
        """Load the object from a file.

        >>> obj = cls.load("obj.json.gz")

        """
        with gzip.open(filename, "rb") as f:
            file_contents = f.read()
            file_contents_decoded = file_contents.decode("utf-8")
            d = json.loads(file_contents_decoded)
            # d = json.loads(f.read().decode("utf-8"))
        return cls.from_dict(d)

    def _post(self):
        """
        Post the object to a pastebin.
        TODO: Implement this method using coop.
        """
        from edsl.utilities.pastebin import post

        post(self)


class RegisterSubclassesMeta(ABCMeta):
    """Metaclass for registering subclasses."""

    _registry = {}

    def __init__(cls, name, bases, nmspc):
        """Register the class in the registry upon creation."""
        super(RegisterSubclassesMeta, cls).__init__(name, bases, nmspc)
        if cls.__name__ != "Base":
            RegisterSubclassesMeta._registry[cls.__name__] = cls

    @staticmethod
    def get_registry():
        """Return the registry of subclasses."""
        return dict(RegisterSubclassesMeta._registry)


class Base(RichPrintingMixin, PersistenceMixin, ABC, metaclass=RegisterSubclassesMeta):
    """Base class for all classes in the package."""

    # def __getitem__(self, key):
    #     return getattr(self, key)

    # @abstractmethod
    # def _repr_html_(self) -> str:
    #     raise NotImplementedError("This method is not implemented yet.")
    
    # @abstractmethod
    # def _repr_(self) -> str:
    #     raise NotImplementedError("This method is not implemented yet.")

    def keys(self):
        """Return the keys of the object."""
        return self.to_dict().keys()
    
    def values(self):
        """Return the values of the object."""
        return self.to_dict().values()

    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html
        return data_to_html(self.to_dict())

    @abstractmethod
    def example():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def rich_print():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def to_dict():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def from_dict():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def code():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    def show_methods(self, show_docstrings=True):
        """Show the methods of the object."""
        public_methods_with_docstrings = [
            (method, getattr(self, method).__doc__)
            for method in dir(self)
            if callable(getattr(self, method)) and not method.startswith("_")
        ]
        if show_docstrings:
            for method, documentation in public_methods_with_docstrings:
                print(f"{method}: {documentation}")
        else:
            return [x[0] for x in public_methods_with_docstrings]



"""Mixin class for exporting results."""
import base64
import csv
import io
from functools import wraps

from typing import Literal, Optional

from edsl.utilities.utilities import is_notebook

from IPython.display import HTML, display
import pandas as pd
from edsl.utilities import (
    print_list_of_dicts_with_rich,
    print_list_of_dicts_as_html_table,
    print_dict_with_rich,
    print_list_of_dicts_as_markdown_table,
)


class ResultsExportMixin:
    """Mixin class for exporting Results objects."""

    def _convert_decorator(func):
        """Convert the Results object to a Dataset object before calling the function."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            """Return the function with the Results object converted to a Dataset object."""
            if self.__class__.__name__ == "Results":
                return func(self.select(), *args, **kwargs)
            elif self.__class__.__name__ == "Dataset":
                return func(self, *args, **kwargs)
            else:
                raise Exception(
                    f"Class {self.__class__.__name__} not recognized as a Results or Dataset object."
                )

        return wrapper

    @_convert_decorator
    def _make_tabular(self, remove_prefix) -> tuple[list, list]:
        """Turn the results into a tabular format."""
        d = {}
        full_header = sorted(list(self.relevant_columns()))
        for entry in self.data:
            key, list_of_values = list(entry.items())[0]
            d[key] = list_of_values
        if remove_prefix:
            header = [h.split(".")[-1] for h in full_header]
        else:
            header = full_header
        num_observations = len(list(self[0].values())[0])
        rows = []
        # rows.append(header)
        for i in range(num_observations):
            row = [d[h][i] for h in full_header]
            rows.append(row)
        return header, rows

    def print_long(self) -> None:
        """Print the results in long format."""
        for result in self:
            if hasattr(result, "combined_dict"):
                d = result.combined_dict
            else:
                d = result
            print_dict_with_rich(d)

    @_convert_decorator
    def print(
        self,
        pretty_labels: Optional[dict] = None,
        filename: Optional[str] = None,
        format: Literal["rich", "html", "markdown"] = None,
        interactive: bool = False,
        split_at_dot: bool = True,
        max_rows = None,
    ) -> None:
        """Print the results in a pretty format.

        :param pretty_labels: A dictionary of pretty labels for the columns.
        :param filename: The filename to save the results to.
        :param format: The format to print the results in. Options are 'rich', 'html', or 'markdown'.
        :param interactive: Whether to print the results interactively in a Jupyter notebook.
        :param split_at_dot: Whether to split the column names at the last dot w/ a newline.

        Example: Print in rich format at the terminal

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.print()
        >>> r.select('how_feeling').print(format = "rich")
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ OK           │
        ├──────────────┤
        │ Great        │
        ├──────────────┤
        │ Terrible     │
        ├──────────────┤
        │ OK           │
        └──────────────┘

        Example: using the pretty_labels parameter

        >>> r.select('how_feeling').print(format="rich", pretty_labels = {'answer.how_feeling': "How you are feeling"})
        ┏━━━━━━━━━━━━━━━━━━━━━┓
        ┃ How you are feeling ┃
        ┡━━━━━━━━━━━━━━━━━━━━━┩
        │ OK                  │
        ├─────────────────────┤
        │ Great               │
        ├─────────────────────┤
        │ Terrible            │
        ├─────────────────────┤
        │ OK                  │
        └─────────────────────┘

        Example: printing in markdown format

        >>> r.select('how_feeling').print(format='markdown')
        | answer.how_feeling |
        |--|
        | OK |
        | Great |
        | Terrible |
        | OK |


        """
        if format is None:
            if is_notebook():
                format = "html"
            else:
                format = "rich"

        if pretty_labels is None:
            pretty_labels = {}

        if format not in ["rich", "html", "markdown"]:
            raise ValueError("format must be one of 'rich', 'html', or 'markdown'.")

        new_data = []
        for index, entry in enumerate(self):
            key, list_of_values = list(entry.items())[0]
            new_data.append({pretty_labels.get(key, key): list_of_values})

        if max_rows is not None:
            for entry in new_data:
                for key in entry:
                    entry[key] = entry[key][:max_rows]

        if format == "rich":
            print_list_of_dicts_with_rich(
                new_data, filename=filename, split_at_dot=split_at_dot
            )
        elif format == "html":
            notebook = is_notebook()
            html = print_list_of_dicts_as_html_table(
                new_data, filename=None, interactive=interactive, notebook=notebook
            )
            #print(html)
            display(HTML(html))
        elif format == "markdown":
            print_list_of_dicts_as_markdown_table(new_data, filename=filename)

    @_convert_decorator
    def to_csv(
        self,
        filename: Optional[str] = None,
        remove_prefix: bool = False,
        download_link: bool = False,
    ):
        """Export the results to a CSV file.

        :param filename: The filename to save the CSV file to.
        :param remove_prefix: Whether to remove the prefix from the column names.
        :param download_link: Whether to display a download link in a Jupyter notebook.

        Example:

        >>> r = create_example_results()
        >>> r.select('how_feeling').to_csv()
        'result.how_feeling\\r\\nBad\\r\\nBad\\r\\nGreat\\r\\nGreat\\r\\n'
        """
        header, rows = self._make_tabular(remove_prefix)

        if filename is not None:
            with open(filename, "w") as f:
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows(rows)
        else:
            output = io.StringIO()
            writer = csv.writer(output)
            writer.writerow(header)
            writer.writerows(rows)

            if download_link:
                csv_file = output.getvalue()
                b64 = base64.b64encode(csv_file.encode()).decode()
                download_link = f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'
                display(HTML(download_link))
            else:
                return output.getvalue()

    @_convert_decorator
    def to_pandas(self, remove_prefix: bool = False) -> pd.DataFrame:
        """Convert the results to a pandas DataFrame.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> r.select('how_feeling').to_pandas()
        answer.how_feeling
        0                 OK
        1              Great
        2           Terrible
        3                 OK

        """
        csv_string = self.to_csv(remove_prefix=remove_prefix)
        csv_buffer = io.StringIO(csv_string)
        df = pd.read_csv(csv_buffer)
        df_sorted = df.sort_index(axis=1)  # Sort columns alphabetically
        return df_sorted
        # return df

    @_convert_decorator
    def to_dicts(self, remove_prefix: bool = False) -> list[dict]:
        """Convert the results to a list of dictionaries.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_dicts()
        [{'answer.how_feeling': 'OK'}, {'answer.how_feeling': 'Great'}, {'answer.how_feeling': 'Terrible'}, {'answer.how_feeling': 'OK'}]

        """
        df = self.to_pandas(remove_prefix=remove_prefix)
        df = df.convert_dtypes()
        list_of_dicts = df.to_dict(orient="records")
        # Convert any pd.NA values to None
        list_of_dicts = [
            {k: (None if pd.isna(v) else v) for k, v in record.items()}
            for record in list_of_dicts
        ]
        return list_of_dicts

    @_convert_decorator
    def to_list(self) -> list[list]:
        """Convert the results to a list of lists.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_list()
        ['OK', 'Great', 'Terrible', 'OK']
        """
        if len(self) == 1:
            return list(self[0].values())[0]
        else:
            return tuple([list(x.values())[0] for x in self])


if __name__ == "__main__":
    import doctest

    doctest.testmod()



class LanguageModelExceptions(Exception):
    pass


class LanguageModelResponseNotJSONError(LanguageModelExceptions):
    pass


class LanguageModelMissingAttributeError(LanguageModelExceptions):
    pass


class LanguageModelAttributeTypeError(LanguageModelExceptions):
    pass


class LanguageModelDoNotAddError(LanguageModelExceptions):
    pass



import json
import os
import requests
from requests.exceptions import ConnectionError
from typing import Any, Optional, Type, Union
from edsl import CONFIG
from edsl.agents import Agent, AgentList
from edsl.questions.QuestionBase import QuestionBase
from edsl.results import Results
from edsl.surveys import Survey
from edsl.data.CacheEntry import CacheEntry


api_url = {
    "development": "http://127.0.0.1:8000",
    "development-testrun": "http://127.0.0.1:8000",
    "production": os.getenv("EXPECTED_PARROT_API_URL"),
}


class Coop:
    """
    Client for the Expected Parrot API.
    """

    def __init__(self, api_key: str = None, run_mode: str = None) -> None:
        """Initialize the client."""
        self.api_key = api_key or os.getenv("EXPECTED_PARROT_API_KEY")
        self.run_mode = run_mode or CONFIG.EDSL_RUN_MODE
        self.run_mode = "production"
        self._api_key_is_valid()

    ################
    # BASIC METHODS
    ################
    def _api_key_is_valid(self) -> None:
        """
        Check if the API key is valid.
        """
        if not self.api_key:
            raise ValueError("API key is required.")
        if not isinstance(self.api_key, str):
            raise ValueError("API key must be a string.")
        response = self._send_server_request(uri="api/v0/validate-apikey", method="GET")
        self._resolve_server_response(response)

    @property
    def headers(self) -> dict:
        """
        Return the headers for the request.
        """
        return {"Authorization": f"Bearer {self.api_key}"}

    @property
    def url(self) -> str:
        """
        Return the URL for the request.
        """
        return api_url[self.run_mode]

    def _send_server_request(
        self,
        uri: str,
        method: str,
        payload: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Send a request to the server and return the response.
        """
        url = f"{self.url}/{uri}"

        try:
            if method.upper() in ["GET", "DELETE"]:
                response = requests.request(
                    method, url, params=params, headers=self.headers
                )
            else:
                response = requests.request(
                    method, url, json=payload, headers=self.headers
                )
        except ConnectionError:
            if self.run_mode == "production":
                raise ConnectionError("Could not connect to the server.")
            else:
                raise ConnectionError(
                    f"\n\n\nCould not connect to the server."
                    f"\nIs the server running? Try:"
                    f"\n   cd ~/coop && make fresh-db && make launch"
                )

        return response

    def _resolve_server_response(self, response: requests.Response) -> None:
        """
        Check the response from the server and raise appropriate errors.
        """
        if response.status_code >= 400:
            raise Exception(response.json().get("detail"))

    ################
    # CACHE METHODS
    ################
    def create_cache_entry(self, cache_entry: CacheEntry) -> dict:
        """
        Create a CacheEntry object.
        """
        response = self._send_server_request(
            uri="api/v0/cache/create-cache-entry",
            method="POST",
            payload={
                "json_string": json.dumps(
                    {"key": cache_entry.key, "value": json.dumps(cache_entry.to_dict())}
                )
            },
        )
        self._resolve_server_response(response)
        return response.json()

    def get_cache_entries(
        self, exclude_keys: Optional[list[str]] = None
    ) -> list[CacheEntry]:
        """
        Return CacheEntry objects from the server.

        :param exclude_keys: exclude CacheEntry objects with these keys.
        """
        if exclude_keys is None:
            exclude_keys = []
        response = self._send_server_request(
            uri="api/v0/cache/get-cache-entries",
            method="POST",
            payload={"json_string": json.dumps(exclude_keys)},
        )
        self._resolve_server_response(response)
        return [
            CacheEntry.from_dict(json.loads(v.get("json_string")))
            for v in response.json()
        ]

    def send_cache_entries(self, cache_entries: dict[str, CacheEntry]) -> None:
        """
        Send a dictionary of CacheEntry objects to the server.
        """
        response = self._send_server_request(
            uri="api/v0/cache/create-cache-entries",
            method="POST",
            payload={
                "json_string": json.dumps(
                    {k: json.dumps(v.to_dict()) for k, v in cache_entries.items()}
                )
            },
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # EDSL METHODS
    ################
    def _edsl_object_to_uri(
        self, object: Union[Type[QuestionBase], Survey, Agent, Results]
    ) -> str:
        """
        Return the URI for the object type.
        """
        if isinstance(object, QuestionBase):
            return "questions"
        elif isinstance(object, Survey):
            return "surveys"
        elif isinstance(object, Agent) or isinstance(object, AgentList):
            return "agents"
        elif isinstance(object, Results):
            return "results"
        else:
            raise ValueError("Incorrect or not supported object type")

    # -----------------
    # A. CREATE METHODS
    # -----------------
    def _create(
        self,
        edsl_object: Union[Type[QuestionBase], Survey, Agent, AgentList, Results],
        public: bool = False,
    ) -> dict:
        """
        Create an EDSL object in the Coop server.

        :param edsl_object: the EDSL object to be sent.
        :param public: whether the object should be public (defaults to False).
        """
        uri = self._edsl_object_to_uri(edsl_object)
        response = self._send_server_request(
            uri=f"api/v0/{uri}",
            method="POST",
            payload={
                "json_string": json.dumps(edsl_object.to_dict()),
                "public": public,
            },
        )
        self._resolve_server_response(response)
        return response.json()

    def create(
        self,
        object: Union[Type[QuestionBase], Survey, Agent, AgentList, Results],
        public: bool = False,
    ):
        """
        Create an EDSL object in the Coop server.
        
        :param object: the EDSL object to be sent.
        :param public: whether the object should be public (defaults to False)
        """
        return self._create(object, public)

    # -----------------
    # B. GET METHODS
    # -----------------
    def _get(self, object_type: str, id: int) -> dict:
        """
        Retrieve an EDSL object from the Coop server.
        """
        response = self._send_server_request(
            uri=f"api/v0/{object_type}/{id}", method="GET"
        )
        self._resolve_server_response(response)
        return json.loads(response.json().get("json_string"))

    def get_question(self, id: int) -> Type[QuestionBase]:
        """
        Retrieve a Question object by its id.
        """
        json_dict = self._get("questions", id)
        return QuestionBase.from_dict(json_dict)

    def get_survey(self, id: int) -> Type[Survey]:
        """
        Retrieve a Survey object by its id.
        """
        json_dict = self._get("surveys", id)
        return Survey.from_dict(json_dict)

    def get_agent(self, id: int) -> Union[Agent, AgentList]:
        """
        Retrieve an Agent or AgentList object by id.
        """
        json_dict = self._get("agents", id)
        if "agent_list" in json_dict:
            return AgentList.from_dict(json_dict)
        else:
            return Agent.from_dict(json_dict)

    def get_results(self, id: int) -> Results:
        """Retrieve a Results object by id."""
        json_dict = self._get("results", id)
        return Results.from_dict(json_dict)

    def get(
        self, object_type: str, id: int
    ) -> Union[Type[QuestionBase], Survey, Agent, AgentList, Results]:
        """
        Retrieve an EDSL object by its id.

        :param object_type: the type of object to retrieve.
        :param id: the id of the object.
        """
        if object_type in {"question", "questions"}:
            return self.get_question(id)
        elif object_type in {"survey", "surveys"}:
            return self.get_survey(id)
        elif object_type in {"agent", "agents"}:
            return self.get_agent(id)
        elif object_type == "results":
            return self.get_results(id)
        else:
            raise ValueError("Object type not recognized")

    def _get_base(self, cls, id):
        """
        Used by the Base class to offer a get functionality.
        """
        if issubclass(cls, QuestionBase):
            return self.get_question(id)
        elif cls == Survey:
            return self.get_survey(id)
        elif cls == Agent or cls == AgentList:
            return self.get_agent(id)
        elif cls == Results:
            return self.get_results(id)
        else:
            raise ValueError("Class type not recognized")

    # -----------------
    # C. GET ALL METHODS
    # -----------------
    @property
    def questions(self) -> list[dict[str, Union[int, QuestionBase]]]:
        """Retrieve all Questions."""
        response = self._send_server_request(uri="api/v0/questions", method="GET")
        self._resolve_server_response(response)
        questions = [
            {
                "id": q.get("id"),
                "question": QuestionBase.from_dict(json.loads(q.get("json_string"))),
            }
            for q in response.json()
        ]
        return questions

    @property
    def surveys(self) -> list[dict[str, Union[int, Survey]]]:
        """Retrieve all Surveys."""
        response = self._send_server_request(uri="api/v0/surveys", method="GET")
        self._resolve_server_response(response)
        surveys = [
            {
                "id": q.get("id"),
                "survey": Survey.from_dict(json.loads(q.get("json_string"))),
            }
            for q in response.json()
        ]
        return surveys

    @property
    def agents(self) -> list[dict[str, Union[int, Agent, AgentList]]]:
        """Retrieve all Agents and AgentLists."""
        response = self._send_server_request(uri="api/v0/agents", method="GET")
        self._resolve_server_response(response)
        agents = []
        for q in response.json():
            agent_dict = json.loads(q.get("json_string"))
            if "agent_list" in agent_dict:
                agent = AgentList.from_dict(agent_dict)
            else:
                agent = Agent.from_dict(agent_dict)
            agents.append({"id": q.get("id"), "agent": agent})
        return agents

    @property
    def results(self) -> list[dict[str, Union[int, Results]]]:
        """Retrieve all Results."""
        response = self._send_server_request(uri="api/v0/results", method="GET")
        self._resolve_server_response(response)
        results = [
            {
                "id": r.get("id"),
                "results": Results.from_dict(json.loads(r.get("json_string"))),
            }
            for r in response.json()
        ]
        return results

    # -----------------
    # D. DELETE METHODS
    # -----------------
    def delete_question(self, id: int) -> dict:
        """Delete a question from the Coop."""
        response = self._send_server_request(
            uri=f"api/v0/questions/{id}", method="DELETE"
        )
        self._resolve_server_response(response)
        return response.json()

    def delete_survey(self, id: int) -> dict:
        """Delete a Survey from the coop."""
        response = self._send_server_request(
            uri=f"api/v0/surveys/{id}", method="DELETE"
        )
        self._resolve_server_response(response)
        return response.json()

    def delete_agent(self, id: int) -> dict:
        """Delete an Agent or AgentList from the coop."""
        response = self._send_server_request(uri=f"api/v0/agents/{id}", method="DELETE")
        self._resolve_server_response(response)
        return response.json()

    def delete_results(self, id: int) -> dict:
        """Delete a Results object from the coop."""
        response = self._send_server_request(
            uri=f"api/v0/results/{id}", method="DELETE"
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # DUNDER METHODS
    ################
    def __repr__(self):
        """Return a string representation of the client."""
        return f"Client(api_key='{self.api_key}', run_mode='{self.run_mode}')"


if __name__ == "__main__":
    from edsl.coop import Coop

    API_KEY = "b"
    RUN_MODE = "development"
    coop = Coop(api_key=API_KEY, run_mode=RUN_MODE)

    # basics
    coop
    coop.headers
    coop.url

    ##############
    # A. QUESTIONS
    ##############
    from edsl.questions import QuestionMultipleChoice
    from edsl.questions import QuestionCheckBox
    from edsl.questions import QuestionFreeText

    # check questions on server (should be an empty list)
    coop.questions
    for question in coop.questions:
        coop.delete_question(question.get("id"))

    # get a question that does not exist (should return None)
    coop.get_question(id=1000)

    # now post a Question
    coop.create(QuestionMultipleChoice.example())
    coop.create(QuestionCheckBox.example(), public=False)
    coop.create(QuestionFreeText.example(), public=True)

    # check all questions
    coop.questions

    # or get question by id
    coop.get_question(id=1)

    # delete the question
    coop.delete_question(id=1)

    # check all questions
    coop.questions

    ##############
    # B. Surveys
    ##############
    from edsl.surveys import Survey

    # check surveys on server (should be an empty list)
    coop.surveys
    for survey in coop.surveys:
        coop.delete_survey(survey.get("id"))

    # get a survey that does not exist (should return None)
    coop.get_survey(id=1)

    # now post a Survey
    coop.create(Survey.example())
    coop.create(Survey.example(), public=False)
    coop.create(Survey.example(), public=True)
    coop.create(Survey(), public=True)
    s = Survey().example()
    for i in range(10):
        q = QuestionFreeText.example()
        q.question_name = f"question_{i}"
        s.add_question(q)
    coop.create(s, public=True)

    # check all surveys
    coop.surveys

    # or get survey by id
    coop.get_survey(id=1)

    # delete the survey
    coop.delete_survey(id=1)

    # check all surveys
    coop.surveys

    ##############
    # C. Agents and AgentLists
    ##############
    from edsl.agents import Agent, AgentList

    # check agents on server (should be an empty list)
    coop.agents
    for agent in coop.agents:
        coop.delete_agent(agent.get("id"))

    # get an agent that does not exist (should return None)
    coop.get_agent(id=2)

    # now post an Agent
    coop.create(Agent.example())
    coop.create(Agent.example(), public=False)
    coop.create(Agent.example(), public=True)
    coop.create(
        Agent(traits={"hair_type": "curly", "skil_color": "white"}), public=True
    )
    coop.create(AgentList.example())
    coop.create(AgentList.example(), public=False)
    coop.create(AgentList.example(), public=True)

    # check all agents
    coop.agents

    # or get agent by id
    coop.get_agent(id=1)

    # delete the agent
    coop.delete_agent(id=1)

    # check all agents
    coop.agents

    ##############
    # D. Results
    ##############
    from edsl.results import Results

    # check results on server (should be an empty list)
    len(coop.results)
    for results in coop.results:
        coop.delete_results(results.get("id"))

    # get a result that does not exist (should return None)
    coop.get_results(id=2)

    # now post a Results
    coop.create(Results.example())
    coop.create(Results.example(), public=False)
    coop.create(Results.example(), public=True)

    # check all results
    coop.results

    # or get results by id
    coop.get_results(id=1)

    # delete the results
    coop.delete_results(id=1)

    # check all results
    coop.results

    ##############
    # E. CACHE
    ##############
    from edsl.data.CacheEntry import CacheEntry

    # should be empty in the beginning
    coop.get_cache_entries()
    # now create one cache entry
    cache_entry = CacheEntry.example()
    coop.create_cache_entry(cache_entry)
    # see that if you try to create it again, you'll get the same id
    coop.create_cache_entry(cache_entry)
    # now get all your cache entries
    coop.get_cache_entries()
    coop.get_cache_entries(exclude_keys=[])
    coop.get_cache_entries(exclude_keys=["a"])
    # this will be empty
    coop.get_cache_entries(exclude_keys=[cache_entry.key])
    # now send many cache entries
    cache_entries = {}
    for i in range(10):
        cache_entry = CacheEntry.example(randomize=True)
        cache_entries[cache_entry.key] = cache_entry
    coop.send_cache_entries(cache_entries)



"""Compose two questions where the answer to q1 is used as an input to q2."""
from edsl.questions import QuestionFunctional
from edsl.questions.QuestionBase import QuestionBase
from edsl.agents import Agent
from edsl.scenarios.Scenario import Scenario


def compose_questions(
    q1: QuestionBase, q2: QuestionBase, question_name: str = None
) -> QuestionFunctional:
    """
    Compose two questions where the answer to q1 is used as an input to q2.

    The resulting question is a question that can be used like other questions.
    Note that the same result can also be achieved in other ways:
    - Using the `add_targeted_memory(q2, q1)` method in Survey
    - Using the __add__ method in Question
    """
    if question_name is None:
        question_name = f"{q1.question_name}_{q2.question_name}"
    if q1.question_name not in q2.question_text:
        raise ValueError(
            f"q2 requires a field not present in q1's answer. "
            f"q1: {q1.question_name}, q2: {q2.question_name}"
        )

    def combo(
        scenario: Scenario, agent_traits: dict[str, str] = None
    ) -> QuestionFunctional:
        """Return the answer to the second question given the answer to the first question."""
        # get the answer to the first question
        first_answer = (
            q1.by(scenario)
            .by(Agent(traits=agent_traits))
            .run()
            .select(q1.question_name)[0]
        )
        # update the scenario with the answer to the first question
        scenario.update({q1.question_name: first_answer})
        # get the answer to the second question
        second_answer = (
            q2.by(scenario)
            .by(Agent(traits=agent_traits))
            .run()
            .select(q2.question_name)[0]
        )
        return second_answer

    return QuestionFunctional(
        question_name=question_name, question_text="functional", func=combo
    )


# UNCOMMENT BELOW TO SEE HOW THIS WORKS

# if __name__ == "__main__":
#     from edsl.questions import QuestionFreeText, QuestionFunctional
#     from edsl.questions.compose_questions import compose_questions
#     from edsl.scenarios.Scenario import Scenario

#     q1 = QuestionFreeText(
#         question_text="What is the capital of {{country}}", question_name="capital"
#     )
#     q2 = QuestionFreeText(
#         question_text="What is the population of {{capital}}",
#         question_name="population",
#     )
#     q3 = compose_questions(q1, q2)

#     jobs = q3.by(
#         Scenario({"country": "France"}),
#         Scenario({"country": "Germany"}),
#         Scenario({"country": "Greece"}),
#     )

#     print("Without an agent persona")
#     results1 = jobs.run()

#     print("Adding an agent persona")
#     results2 = jobs.by(
#         Agent(traits={"name": "Bob, who always mentions his travel agency."})
#     ).run()
#     results2.select("capital_population").table()

#     q1 = QuestionFreeText(
#         question_text="What is the capital of {{country}}", question_name="capital"
#     )
#     q2 = QuestionFreeText(
#         question_text="Is this {{population}} large?", question_name="population"
#     )
#     q3 = compose_questions(q1, q2)
#     print("Should throw an exception:")
#     try:
#         q3.by(Scenario({"country": "France"})).run()
#     except ValueError as e:
#         print(e)



"""This module contains the Interview class, which is responsible for conducting an interview asynchronously."""
from __future__ import annotations
import traceback
import asyncio
import time
from typing import Any, Type, List, Generator

from edsl.agents import Agent
from edsl.language_models import LanguageModel
from edsl.scenarios import Scenario
from edsl.surveys import Survey

from edsl.jobs.Answers import Answers
from edsl.surveys.base import EndOfSurvey
from edsl.jobs.buckets.ModelBuckets import ModelBuckets

from edsl.jobs.tasks.TaskCreators import TaskCreators

from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog
from edsl.jobs.interviews.interview_exception_tracking import (
    InterviewExceptionCollection,
    InterviewExceptionEntry,
)
from edsl.jobs.interviews.retry_management import retry_strategy
from edsl.jobs.interviews.InterviewTaskBuildingMixin import InterviewTaskBuildingMixin
from edsl.jobs.interviews.InterviewStatusMixin import InterviewStatusMixin


class Interview(InterviewStatusMixin, InterviewTaskBuildingMixin):
    """
    An 'interview' is one agent answering one survey, with one language model, for a given scenario.

    The main method is `async_conduct_interview`, which conducts the interview asynchronously.
    """

    def __init__(
        self,
        agent: Agent,
        survey: Survey,
        scenario: Scenario,
        model: Type[LanguageModel],
        debug: bool = False,
        iteration: int = 0,
        cache=None,
        sidecar_model=None,
    ):
        """Initialize the Interview instance.

        :param agent: the agent being interviewed.
        :param survey: the survey being administered to the agent.
        :param scenario: the scenario that populates the survey questions.
        :param model: the language model used to answer the questions.

        """
        self.agent = agent
        self.survey = survey
        self.scenario = scenario
        self.model = model
        self.debug = debug
        self.iteration = iteration
        self.cache = cache
        # will get filled in as interview progresses
        self.answers: dict[str, str] = Answers()

        # Trackers
        self.task_creators = TaskCreators()  # tracks the task creators
        self.exceptions = InterviewExceptionCollection()
        self._task_status_log_dict = InterviewStatusLog()

        # dictionary mapping question names to their index in the survey."""
        self.to_index = {
            question_name: index
            for index, question_name in enumerate(self.survey.question_names)
        }

    async def async_conduct_interview(
        self,
        *,
        model_buckets: ModelBuckets = None,
        debug: bool = False,
        stop_on_exception: bool = False,
        sidecar_model=None,
    ) -> tuple["Answers", List[dict[str, Any]]]:
        """
        Conduct an Interview asynchronously.

        :param model_buckets: a dictionary of token buckets for the model.
        :param debug: run without calls to LLM.
        :param stop_on_exception: if True, stops the interview if an exception is raised.

        Example usage:

        >>> i = Interview.example()
        >>> answers = asyncio.run(i.async_conduct_interview())
        >>> answers['q0']
        'yes'

        """
        self.sidecar_model = sidecar_model
        # if no model bucket is passed, create an 'infinity' bucket with no rate limits
        model_buckets = model_buckets or ModelBuckets.infinity_bucket()
        # build the tasks using the InterviewTaskBuildingMixin
        self.tasks = self._build_question_tasks(
            debug=debug, model_buckets=model_buckets
        )
        # 'Invigilators' are used to administer the survey
        self.invigilators = list(self._build_invigilators(debug=debug))
        # await the tasks being conducted
        await asyncio.gather(*self.tasks, return_exceptions=not stop_on_exception)
        self.answers.replace_missing_answers_with_none(self.survey)
        valid_results = list(self._extract_valid_results())
        return self.answers, valid_results

    def _extract_valid_results(
        self, print_traceback=False
    ) -> Generator["Answers", None, None]:
        """Extract the valid results from the list of results.

        :param print_traceback: if True, print the traceback of any exceptions.
        """
        # we only need to print the warning once if a task failed.
        # warning_printed = False
        # warning_header = textwrap.dedent(
        #     """\
        #     WARNING: At least one question in the survey was not answered.
        #     """
        # )
        # # there should be one one invigilator for each task
        assert len(self.tasks) == len(self.invigilators)

        for task, invigilator in zip(self.tasks, self.invigilators):
            if task.done():
                try:  # task worked
                    result = task.result()
                except asyncio.CancelledError as e:  # task was cancelled
                    result = invigilator.get_failed_task_result()

                # We don't want to log cancelled tasks, as this is expected behavior
                ## TODO: Currently, we only log errors at the question-answering phase
                ## Do we want to log exceptions here as well?
                #     exception_entry = InterviewExceptionEntry(
                #         exception=repr(e),
                #         time=time.time(),
                #         traceback=traceback.format_exc(),
                #     )
                #     self.exceptions.add(task.edsl_name, exception_entry)
                except Exception as e:  # any other kind of exception in the task
                    exception_entry = InterviewExceptionEntry(
                        exception=repr(e),
                        time=time.time(),
                        traceback=traceback.format_exc(),
                    )
                    self.exceptions.add(task.edsl_name, exception_entry)
                    # if not warning_printed:
                    #     warning_printed = True
                    #     print(warning_header)

                    error_message = f"Task `{task.edsl_name}` failed with `{e.__class__.__name__}`:`{e}`."
                    # print(error_message)
                    # if print_traceback:
                    #    traceback.print_exc()
                    result = invigilator.get_failed_task_result()

                yield result
            else:
                raise ValueError(f"Task {task.edsl_name} is not done.")

    #######################
    # Dunder methods
    #######################
    def __repr__(self) -> str:
        """Return a string representation of the Interview instance."""
        return f"Interview(agent = {self.agent}, survey = {self.survey}, scenario = {self.scenario}, model = {self.model})"


if __name__ == "__main__":
    """Test the Interview class."""
    from edsl.language_models import LanguageModelOpenAIThreeFiveTurbo
    from edsl.agents import Agent
    from edsl.surveys import Survey
    from edsl.scenarios import Scenario
    from edsl.questions import QuestionMultipleChoice

    # from edsl.jobs.Interview import Interview

    #  a survey with skip logic
    q0 = QuestionMultipleChoice(
        question_text="Do you like school?",
        question_options=["yes", "no"],
        question_name="q0",
    )
    q1 = QuestionMultipleChoice(
        question_text="Why not?",
        question_options=["killer bees in cafeteria", "other"],
        question_name="q1",
    )
    q2 = QuestionMultipleChoice(
        question_text="Why?",
        question_options=["**lack*** of killer bees in cafeteria", "other"],
        question_name="q2",
    )
    s = Survey(questions=[q0, q1, q2])
    s = s.add_rule(q0, "q0 == 'yes'", q2)

    # create an interview
    a = Agent(traits=None)

    def direct_question_answering_method(self, question, scenario):
        """Answer a question directly."""
        raise Exception("Error!")
        # return "yes"

    a.add_direct_question_answering_method(direct_question_answering_method)
    scenario = Scenario()
    m = LanguageModelOpenAIThreeFiveTurbo()
    I = Interview(agent=a, survey=s, scenario=scenario, model=m)

    result = asyncio.run(I.async_conduct_interview())
    # # conduct five interviews
    # for _ in range(5):
    #     I.conduct_interview(debug=True)

    # # replace missing answers
    # I
    # repr(I)
    # eval(repr(I))
    # print(I.task_status_logs.status_matrix(20))
    status_matrix = I.task_status_logs.status_matrix(20)
    numerical_matrix = I.task_status_logs.numerical_matrix(20)
    I.task_status_logs.visualize()

    I.exceptions.print()
    I.exceptions.ascii_table()



import asyncio
from enum import Enum
from typing import Literal, List, Type, DefaultDict
from collections import UserDict, defaultdict

from edsl.jobs.tasks.task_management import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.jobs.pricing import pricing, TokenPricing
from edsl.jobs.tasks.task_status_enum import TaskStatus

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]


class InterviewStatistic(UserDict):
    @staticmethod
    def _format_number(number, digits=0, units=""):
        """Format a number.

        :param number: the number to format
        :param digits: the number of digits to display
        :param units: the units to display

        Example usage:

        >>> InterviewStatistic._format_number(1000, 1, "sec.")
        '1,000.0 sec.'
        """
        if type(number) == str:
            return number
        else:
            return f"{number:,.{digits}f}" + " " + units

    @property
    def _pretty_name(self):
        """Return a pretty name for the statistic.

        Example usage:

        >>> InterviewStatistic("elapsed_time", value=100, digits=1, units="sec.").pretty_name
        'Elapsed time'
        """
        return self.name.replace("_", " ").capitalize()

    def __init__(
        self,
        name: str,
        value: float,
        digits: int = 0,
        units: str = "",
        pretty_name: str = None,
    ):
        """Create a new InterviewStatistic object."""
        self.name = name
        self.value = value
        self.digits = digits
        self.units = units
        self.pretty_name = pretty_name or self._pretty_name

        super().__init__(
            {self.pretty_name: self._format_number(self.value, self.digits, self.units)}
        )

        self.raw: dict = {self.name: self.value}


class InterviewStatisticsCollection(UserDict):
    """A collection of interview statistics."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.raw: dict = {}

    def add_stat(self, statistic: InterviewStatistic):
        """Add a statistic to the collection.

        Each statistic is a dictionary with a single key-value pair.

        Example usage:

        >>> isc = InterviewStatisticsCollection()
        >>> isc.add_stat(InterviewStatistic("elapsed_time", value=100, digits=1, units="sec."))
        >>> isc.raw
        {'elapsed_time': 100}
        """
        self.update(statistic)
        self.raw.update(statistic.raw)


def enum_converter(obj):
    if isinstance(obj, Enum):
        return obj.name  # or obj.value if you prefer the enum's value
    raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")


if __name__ == "__main__":
    # pass
    import doctest

    doctest.testmod()
    # Create a JobsRunHistory object
    # jrh = JobsRunHistory()

    # # Add some data to it
    # jrh.append({"elapsed_time": 0, "completed_interviews": 0})
    # jrh.append({"elapsed_time": 1, "completed_interviews": 1})
    # jrh.append({"elapsed_time": 2, "completed_interviews": 2})

    # # Save the data to a file
    # jrh.to_json("jobs_run_history.json")

    # # Read the data from the file
    # jrh2 = JobsRunHistory.from_json("jobs_run_history.json")

    # # Plot the data
    # jrh2.plot_completion_times()



from collections import UserList


class TasksList(UserList):
    def status(self, debug=False):
        if debug:
            for task in self:
                print(f"Task {task.edsl_name}")
                print(f"\t DEPENDS ON: {task.depends_on}")
                print(f"\t DONE: {task.done()}")
                print(f"\t CANCELLED: {task.cancelled()}")
                if not task.cancelled():
                    if task.done():
                        print(f"\t RESULT: {task.result()}")
                    else:
                        print(f"\t RESULT: None - Not done yet")

            print("---------------------")
